"""Create morphologies from the Steiner Tree solutions."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import logging
from pathlib import Path

import numpy as np
import pandas as pd
from morphio import PointLevel
from morphio import SectionType
from neurom.core import Morphology
from plotly.subplots import make_subplots
from plotly_helper.neuron_viewer import NeuronBuilder

from axon_synthesis.constants import FROM_COORDS_COLS
from axon_synthesis.constants import TO_COORDS_COLS
from axon_synthesis.typing import FileType
from axon_synthesis.utils import add_camera_sync
from axon_synthesis.utils import build_layout_properties
from axon_synthesis.utils import get_morph_pts
from axon_synthesis.utils import save_morphology
from axon_synthesis.utils import sublogger


def plot(morph, initial_morph, figure_path):
    """Plot the Steiner morphology."""
    morph_name = morph.name

    fig = make_subplots(
        cols=2,
        specs=[[{"is_3d": True}, {"is_3d": True}]],
        subplot_titles=["Main trunk", "Initial morphology"],
    )

    # Build the generated figure
    gen_builder = NeuronBuilder(morph, "3d", line_width=4, title=f"{morph_name}")
    gen_fig = gen_builder.get_figure()["data"]
    fig.add_traces(
        gen_fig,
        rows=[1] * len(gen_fig),
        cols=[1] * len(gen_fig),
    )

    # Build the initial figure
    if initial_morph.sections:
        initial_builder = NeuronBuilder(initial_morph, "3d", line_width=4, title=f"{morph_name}")
        initial_fig = initial_builder.get_figure()["data"]
        fig.add_traces(
            initial_fig,
            rows=[1] * len(initial_fig),
            cols=[2] * len(initial_fig),
        )

    layout_props = build_layout_properties(
        np.vstack([morph.points, get_morph_pts(initial_morph)]), 0.1
    )

    fig.update_scenes(layout_props)
    fig.update_layout(title=morph_name)

    Path(figure_path).parent.mkdir(parents=True, exist_ok=True)
    fig.write_html(figure_path)

    # Update the HTML file to synchronize the cameras between the two plots
    add_camera_sync(figure_path)


def build_and_graft_trunk(
    morph: Morphology,
    source_section_id: int,
    # nodes: pd.DataFrame,
    edges: pd.DataFrame,
    *,
    output_path: FileType | None = None,
    figure_path: FileType | None = None,
    initial_morph: Morphology | None = None,
    logger: logging.Logger | logging.LoggerAdapter | None = None,
) -> int:
    """Build and graft a trunk to a morphology from a set of nodes and edges."""
    logger = sublogger(logger, __name__)

    # Build the synthesized axon
    active_sections = []

    edges["section_id"] = -999
    edges["reversed_edge"] = False

    edges_tmp = edges[
        [
            "from",
            "to",
            *FROM_COORDS_COLS,
            *TO_COORDS_COLS,
            "source_is_terminal",
            "target_is_terminal",
        ]
    ].copy()

    if source_section_id == -1:
        # Create a root section to start a new axon
        roots = edges_tmp.loc[edges_tmp["from"] == 0]
        if len(roots) > 1:
            # Handle bifurcation at root
            from_pt = roots[FROM_COORDS_COLS].to_numpy()[0]
            to_pt = np.concatenate([[from_pt], roots[TO_COORDS_COLS].to_numpy()]).mean(axis=0)
            # edges.loc[roots.index, FROM_COORDS_COLS] = [to_pt] * len(roots)
            edges_tmp.loc[roots.index, FROM_COORDS_COLS] = [to_pt] * len(roots)
            target_idx = 0
            roots_is_terminal = False
        else:
            from_pt = roots[FROM_COORDS_COLS].to_numpy()[0]
            to_pt = roots[TO_COORDS_COLS].to_numpy()[0]
            target_idx = roots["to"].to_numpy()[0]
            roots_is_terminal = bool(roots["target_is_terminal"].to_numpy()[0])

            # Remove the root edge
            edges_tmp = edges_tmp.drop(roots.index)

        # Build the root section
        root_section = morph.append_root_section(
            PointLevel(
                [
                    from_pt,
                    to_pt,
                ],
                [0, 0],
            ),
            SectionType.axon,
        )
        if roots_is_terminal:
            edges.loc[roots.index, "section_id"] = root_section.id
    else:
        # Attach the axon to the grafting section
        root_section = morph.section(source_section_id)
        edges.loc[edges["from"] == 0, "section_id"] = root_section.id
        target_idx = 0

        # Check that the source point is consistent with the last section point
        shifts = edges_tmp.loc[edges_tmp["from"] == 0, FROM_COORDS_COLS] - root_section.points[-1]
        if (~np.isclose(shifts, 0)).any():
            logger.warning(
                "The source points (%s) are not all equal to the parent section point (%s) and "
                "are thus shifted",
                edges_tmp.loc[edges_tmp["from"] == 0, FROM_COORDS_COLS].to_numpy().tolist(),
                root_section.points[-1].tolist(),
            )
            edges_tmp.loc[edges_tmp["from"] == 0, FROM_COORDS_COLS] -= shifts

    active_sections.append((root_section, target_idx))

    while active_sections:
        current_section, target = active_sections.pop()
        already_added = []
        for label, row in edges_tmp.loc[edges_tmp["from"] == target].iterrows():
            already_added.append(label)
            new_section = current_section.append_section(
                PointLevel(
                    [
                        row[FROM_COORDS_COLS].to_numpy(),
                        row[TO_COORDS_COLS].to_numpy(),
                    ],
                    [0, 0],
                ),
                SectionType.axon,
            )
            active_sections.append(
                (
                    new_section,
                    row["to"],
                ),
            )
            edges.loc[label, "section_id"] = new_section.id  # type: ignore[index]
        for label, row in edges_tmp.loc[edges_tmp["to"] == target].iterrows():
            already_added.append(label)
            new_section = current_section.append_section(
                PointLevel(
                    [
                        row[TO_COORDS_COLS].to_numpy(),
                        row[FROM_COORDS_COLS].to_numpy(),
                    ],
                    [0, 0],
                ),
                SectionType.axon,
            )
            active_sections.append(
                (
                    new_section,
                    row["from"],
                ),
            )
            edges.loc[label, ["section_id", "reversed_edge"]] = [new_section.id, True]  # type: ignore[index]
        edges_tmp = edges_tmp.drop(already_added)

    # At this point we do not merge consecutive sections that are not separated by a
    # bifurcation, we do it only at the very end of the process. This is to keep section
    # IDs synced with point IDs.

    if output_path is not None:
        # Export the morphology
        save_morphology(morph, output_path, msg=f"Export trunk morphology to {output_path}")

    if figure_path is not None:
        logger.info("Export trunk figure to %s", figure_path)
        plot(morph, initial_morph, figure_path)

    return root_section.id



"""Some utils for the AxonSynthesis package."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import collections.abc
import hashlib
import inspect
import json
import logging
import re
import shutil
import tempfile
import warnings
from collections.abc import MutableMapping
from contextlib import contextmanager
from copy import deepcopy
from pathlib import Path

try:
    from mpi4py import MPI

    mpi_enabled = True
except ImportError:
    mpi_enabled = False

try:
    logging.getLogger("tecio").disabled = True
    from construct import Float32l as Float
    from construct import GreedyRange
    from tecio import TecDatasetAux
    from tecio import TecHeader
    from tecio import TecplotMarker
    from tecio import ZoneType
    from tecio import gen_data_struct
    from tecio import gen_zone_struct

    has_tecio = True
except ImportError:
    has_tecio = False


import dask.dataframe as dd
import networkx as nx
import numpy as np
import pandas as pd
from attrs import define
from attrs import field
from attrs import validators
from bluepyparallel import evaluate
from bluepyparallel import init_parallel_factory
from dask.distributed import LocalCluster
from morph_tool.converter import convert
from morph_tool.utils import is_morphology
from morphio.mut import Morphology as MorphIoMorphology
from neurom import NeuriteType
from neurom import load_morphology as neurom_load_morphology
from neurom.core import Morphology
from neurom.core.morphology import iter_neurites
from neurom.core.soma import SomaType
from neurom.geom.transform import Translation
from voxcell.cell_collection import CellCollection

from axon_synthesis.constants import COORDS_COLS
from axon_synthesis.constants import FROM_COORDS_COLS
from axon_synthesis.constants import TO_COORDS_COLS
from axon_synthesis.typing import FileType
from axon_synthesis.typing import RegionIdsType
from axon_synthesis.typing import SeedType

_DISTRIBUTED_LOGGERS = [
    "asyncio",
    "distributed",
    "distributed.batched",
    "distributed.core",
    "distributed.http",
    "distributed.http.proxy",
    "distributed.nanny",
    "distributed.scheduler",
    "distributed.worker",
]
LOGGER = logging.getLogger(__name__)


@define
class ParallelConfig:
    """Class to store the parallel configuration.

    Attributes:
        nb_processes: The number of processes.
        dask_config: The dask configuration to use.
        progress_bar: If set to True, a progress bar is displayed during computation.
        use_mpi: Trigger the use of MPI.
    """

    nb_processes: int = field(default=0, validator=validators.ge(0))
    dask_config: dict | None = field(default=None)
    progress_bar: bool = field(default=False)
    use_mpi: bool = field(default=False)


class MorphNameAdapter(logging.LoggerAdapter):
    """Add the morphology name and optionally the axon ID to the log entries."""

    def process(
        self, msg: str, kwargs: MutableMapping[str, object]
    ) -> tuple[str, MutableMapping[str, object]]:
        """Add extra information to the log entry."""
        if self.extra is not None:
            header = f"morphology {self.extra['morph_name']}"
            if "axon_id" in self.extra:
                header += f" (axon {self.extra['axon_id']})"
            return f"{header}: {msg}", kwargs
        return "", kwargs


def sublogger(
    logger: logging.Logger | logging.LoggerAdapter | None, name: str
) -> logging.Logger | logging.LoggerAdapter:
    """Get a sub-logger with specific name."""
    if logger is not None:
        new_logger = logger.manager.getLogger(name)
        if isinstance(logger, logging.LoggerAdapter):
            return logger.__class__(new_logger, logger.extra)
        return new_logger
    return logging.getLogger(name)


def setup_logger(level: str = "info", prefix: str = "", suffix: str = ""):
    """Setup application logger."""
    if mpi_enabled:  # pragma: no cover
        comm = MPI.COMM_WORLD  # pylint: disable=c-extension-no-member
        if comm.Get_size() > 1:
            rank = comm.Get_rank()
            prefix = f"#{rank} - {prefix}"
    level = level.lower()
    levels = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warning": logging.WARNING,
        "error": logging.ERROR,
        "critical": logging.CRITICAL,
    }
    logging.basicConfig(
        format=prefix + "%(asctime)s - %(name)s - %(levelname)s - %(message)s" + suffix,
        datefmt="%Y-%m-%dT%H:%M:%S",
        level=levels[level],
    )

    if levels[level] >= logging.INFO:  # pragma: no cover
        for logger_name in [
            "distributed",
            "h5py",
            "morph_tool.converter",
        ]:
            logging.getLogger(logger_name).level = max(
                logging.getLogger(logger_name).level,
                logging.WARNING,
            )


class TemporaryDirectory(tempfile.TemporaryDirectory):
    """Backport of delete parameter for TemporaryDirectory."""

    def __init__(self, *args, delete=True, **kwargs):
        """Constructor of TemporaryDirectory."""
        super().__init__(*args, **kwargs)
        if not delete:
            self._finalizer.detach()  # type: ignore[attr-defined]


class CleanableDirectory:
    """A class to create a directory that can be cleaned up."""

    def __init__(self, dir_name, *, parents=True, exist_ok=False):
        """Constructor of the CleanableDirectory."""
        dir_name = Path(dir_name)
        dir_name.mkdir(parents=parents, exist_ok=exist_ok)
        self.name = str(dir_name)

    def cleanup(self):
        """Clean up the directory."""
        shutil.rmtree(self.name)


def temp_dir(*args, **kwargs):
    """Create a temporary directory."""
    return TemporaryDirectory(*args, **kwargs)


def fill_diag(mat, val=1):
    """Fill the diagonal of the given matrix."""
    np.fill_diagonal(mat, val)
    return mat


def cols_to_json(df, cols):
    """Transform the given columns from Python objects to JSON strings."""
    df = df.copy(deep=False)  # noqa: PD901
    for col in cols:
        df[col] = df[col].map(json.dumps)
    return df


def cols_from_json(df, cols):
    """Transform the given columns to Python objects from JSON strings."""
    df = df.copy(deep=False)  # noqa: PD901
    for col in cols:
        df[col] = df[col].map(json.loads)
    return df


def get_layers(atlas, pos):
    """Get layer data."""
    # TODO: get layer from the region names?
    names, ids = atlas.get_layers()
    layers = np.zeros_like(atlas.brain_regions.raw, dtype="uint8")
    layer_mapping = {}
    for layer_id, (ids_set, layer) in enumerate(zip(ids, names)):
        layer_mapping[layer_id] = layer
        layers[np.isin(atlas.brain_regions.raw, list(ids_set))] = layer_id + 1
    layer_field = atlas.brain_regions.with_data(layers)
    return layer_field.lookup(pos, outer_value=0)


def add_camera_sync(fig_path):
    """Update the HTML file to synchronize the cameras between the two plots."""
    with Path(fig_path).open(encoding="utf-8") as f:
        tmp = f.read()
        id_match = re.match('.*id="([^ ]*)" .*', tmp, flags=re.DOTALL)
        if id_match is None:
            msg = f"Could not find the figure ID in {fig_path}"
            raise ValueError(msg)
        fig_id = id_match.group(1)

    js = f"""
    <script>
    var gd = document.getElementById('{fig_id}');
    var isUnderRelayout = false

    gd.on('plotly_relayout', () => {{
      console.log('relayout', isUnderRelayout)
      if (!isUnderRelayout) {{
        Plotly.relayout(gd, 'scene2.camera', gd.layout.scene.camera)
          .then(() => {{ isUnderRelayout = false }}  )
      }}

      isUnderRelayout = true;
    }})
    </script>
    """

    with Path(fig_path).open("w", encoding="utf-8") as f:
        f.write(tmp.replace("</body>", js + "</body>"))


def get_nested_morphology_paths(morph_dir, *, max_level=None):
    """Get all morphology paths from a given directory and sub directories."""
    morph_dir = Path(morph_dir)
    morphology_paths = []
    for morph_path in morph_dir.iterdir():
        if morph_path.is_dir() and (max_level is None or max_level > 0):
            morphology_paths.extend(get_nested_morphology_paths(morph_path))
        elif is_morphology(morph_path):
            morphology_paths.append(morph_path)
    return morphology_paths


def get_morphology_paths(morph_dir, *, max_level=None):
    """Get all morphology paths from a given directory."""
    morphology_paths = get_nested_morphology_paths(morph_dir, max_level=max_level)
    paths = pd.DataFrame(morphology_paths, columns=["morph_path"])
    paths.loc[:, "morph_name"] = paths["morph_path"].apply(
        lambda x: str(x.relative_to(morph_dir).with_suffix(""))
    )
    paths.loc[:, "morph_path"] = paths["morph_path"].apply(str)
    return paths


def get_axons(morph, axon_ids=None):
    """Get axons of the given morphology."""
    axons = [i for i in morph.neurites if i.type == NeuriteType.axon]
    if axon_ids is None:
        return axons
    if isinstance(axon_ids, int):
        return axons[axon_ids]
    return [axons[i] for i in axon_ids]


def keep_only_neurites(morph, neurite_type=None, neurite_idx=None, *, copy=False):
    """Delete neurites except the ones with a given type or index.

    .. note::
        If both the type and index are given, the index is only applied for the neurites of the
        given type.
    """
    if copy:
        morph = Morphology(morph)

    if neurite_type is not None:
        for i in morph.root_sections:
            if i.type != neurite_type:
                morph.delete_section(i)

    if neurite_idx is not None:
        to_delete = []
        for num, i in enumerate(morph.root_sections):
            if num != neurite_idx:
                to_delete.append(i)
        for i in to_delete:
            morph.delete_section(i)

    return morph


def neurite_to_pts(neurite, *, keep_section_segments=False, edges_with_coords=False):
    """Extract points and segments from a neurite."""
    graph_nodes = []
    graph_edges = []
    node_id = -1
    last_pts = {None: -1}
    for section in neurite.iter_sections():
        is_terminal = not bool(section.children)

        if section.parent is None:
            # Add first point of the root section
            graph_nodes.append((node_id, *section.points[0, :4], True, -1, 0))
            last_pt = last_pts[None]
        else:
            last_pt = last_pts[section.parent.id]

        # Add segment points
        pts = section.points[1:, :4] if keep_section_segments else section.points[-1:, :4]
        len_pts = len(pts) - 1

        for num, i in enumerate(pts.tolist()):
            node_id = node_id + 1
            graph_nodes.append((node_id, *i, num == len_pts and is_terminal, section.id, num))
            graph_edges.append((last_pt, node_id))
            last_pt = node_id

        last_pts[section.id] = last_pt

    nodes = pd.DataFrame(
        graph_nodes,
        columns=["id", *COORDS_COLS, "radius", "is_terminal", "section_id", "sub_segment_num"],
    )
    nodes.set_index("id", inplace=True)

    edges = pd.DataFrame(graph_edges, columns=["source", "target"])
    edges = edges.sort_values(
        ["source", "target"],
    ).reset_index(drop=True)

    if edges_with_coords:
        edges = edges.merge(nodes, left_on="source", right_index=True).rename(
            columns={
                COORDS_COLS.X: FROM_COORDS_COLS.X,
                COORDS_COLS.Y: FROM_COORDS_COLS.Y,
                COORDS_COLS.Z: FROM_COORDS_COLS.Z,
            }
        )
        edges = edges.merge(
            nodes, left_on="target", right_index=True, suffixes=("_from", "_to")
        ).rename(
            columns={
                COORDS_COLS.X: TO_COORDS_COLS.X,
                COORDS_COLS.Y: TO_COORDS_COLS.Y,
                COORDS_COLS.Z: TO_COORDS_COLS.Z,
            }
        )

    return nodes, edges


def neurite_to_graph(neurite, graph_cls=nx.DiGraph, *, keep_section_segments=False, **graph_kwargs):
    """Transform a neurite into a graph."""
    nodes, edges = neurite_to_pts(neurite, keep_section_segments=keep_section_segments)
    graph = nx.from_pandas_edgelist(edges, create_using=graph_cls, **graph_kwargs)
    nx.set_node_attributes(
        graph, nodes[["section_id", *COORDS_COLS, "radius", "is_terminal"]].to_dict("index")
    )

    return nodes, edges, graph


def neurite_to_graph_old(neurite, graph_cls=nx.DiGraph, **graph_kwargs):
    """Transform a neurite into a graph."""
    graph_nodes = []
    graph_edges = []
    for section in neurite.iter_sections():
        is_terminal = not bool(section.children)
        if section.parent is None:
            graph_nodes.append((-1, *section.points[0, :3], True))
            graph_edges.append((-1, section.id))

        graph_nodes.append((section.id, *section.points[-1, :3], is_terminal))

        graph_edges.extend((section.id, child.id) for child in section.children)

    nodes = pd.DataFrame(graph_nodes, columns=["id", *COORDS_COLS, "is_terminal"])
    nodes.set_index("id", inplace=True)

    edges = pd.DataFrame(graph_edges, columns=["source", "target"])
    graph = nx.from_pandas_edgelist(edges, create_using=graph_cls, **graph_kwargs)
    nx.set_node_attributes(graph, nodes[[*COORDS_COLS, "is_terminal"]].to_dict("index"))

    return nodes, edges, graph


def export_morph_edges(morph, output_path, neurite_filter=None, logger=None):
    """Export the morphology as DataFrame for later analysis."""
    if logger is None:
        logger = LOGGER
    all_edges = []
    for i in iter_neurites(morph, filt=neurite_filter):
        _nodes, edges = neurite_to_pts(i, keep_section_segments=True, edges_with_coords=True)
        edges.loc[:, "neurite_type"] = i.type
        all_edges.append(edges)
    if all_edges:
        edges_df = pd.concat(all_edges, ignore_index=True)
        Path(output_path).parent.mkdir(parents=True, exist_ok=True)
        edges_df.to_csv(output_path, index=False)
        logger.debug("Exported morphology edges to '%s'", str(output_path))
    else:
        logger.debug("No edge to export to '%s'", str(output_path))


@contextmanager
def disable_loggers(*logger_names):
    """A context manager that will disable logging messages triggered during the body.

    Args:
        *logger_names (str): The names of the loggers to be disabled.
    """
    loggers = (
        [logging.getLogger()] if not logger_names else [logging.getLogger(i) for i in logger_names]
    )

    disabled_loggers = [(i, i.disabled) for i in loggers]

    try:
        for i, _ in disabled_loggers:
            i.disabled = True
        yield
    finally:
        for i, j in disabled_loggers:
            i.disabled = j


@contextmanager
def disable_distributed_loggers():
    """A context manager that will disable logging messages from the 'distributed' library."""
    with disable_loggers(*_DISTRIBUTED_LOGGERS):
        yield


def permanently_disable_distributed_loggers():
    """Permanently disable logging messages from the 'distributed' library."""
    loggers = [logging.getLogger(i) for i in _DISTRIBUTED_LOGGERS]

    for i in loggers:
        i.disabled = True


@contextmanager
def ignore_warnings(*ignored_warnings):
    """A context manager that will ignore warnings raised during the body.

    Args:
        *ignored_warnings (Warning): The classes of the warnings to be ignored.
    """
    with warnings.catch_warnings():
        for i in ignored_warnings:
            warnings.filterwarnings("ignore", category=i)
        yield


def recursive_to_str(data):
    """Cast all Path objects into str objects in a given dict."""
    new_data = deepcopy(data)
    for k, v in new_data.items():
        if isinstance(v, dict):
            new_data[k] = recursive_to_str(v)
        elif isinstance(v, Path):
            new_data[k] = str(v)
    return new_data


def recursive_update(data, updates):
    """Update a dictionary with another with nested values."""
    for k, v in updates.items():
        if isinstance(v, collections.abc.Mapping):
            data[k] = recursive_update(data.get(k, {}), v)
        else:
            data[k] = v
    return data


def merge_json_files(*files):
    """Merge several JSON files together.

    The order is important: the files will be updated by all the next files in the list.
    """
    result: dict = {}
    for i in files:
        file = Path(i)
        if file.exists():
            with file.open(encoding="utf-8") as f:
                recursive_update(result, json.load(f))
    return result


def compute_bbox(points, relative_buffer=None, absolute_buffer=None):
    """Compute the bounding box of the given points and optionally apply a buffer to it."""
    bbox = np.vstack([points.min(axis=0), points.max(axis=0)])
    bbox_buffer = None
    if relative_buffer is not None:
        bbox_buffer = (bbox[1] - bbox[0]) * relative_buffer
    if absolute_buffer is not None and (bbox_buffer is None or absolute_buffer > bbox_buffer):
        bbox_buffer = absolute_buffer
    if bbox_buffer is not None:
        bbox[0] -= bbox_buffer
        bbox[1] += bbox_buffer
    return bbox


def compute_aspect_ratios(bbox):
    """Compute the aspect ratios of a bounding box."""
    aspect_ratios = bbox[1] - bbox[0]
    aspect_ratios /= aspect_ratios[np.argmax(aspect_ratios)]
    return aspect_ratios


def build_layout_properties(
    pts, relative_buffer: float | None = None, absolute_buffer: float | None = None
) -> dict:
    """Build a dictionary with layout properties for Plotly figures."""
    bbox = compute_bbox(pts, relative_buffer, absolute_buffer)
    aspect_ratios = compute_aspect_ratios(bbox)

    return {
        "aspectmode": "manual",
        "aspectratio": {"x": aspect_ratios[0], "y": aspect_ratios[1], "z": aspect_ratios[2]},
        "xaxis": {
            "range": bbox[:, 0],
        },
        "yaxis": {
            "range": bbox[:, 1],
        },
        "zaxis": {
            "range": bbox[:, 2],
        },
    }


def get_code_location(back_frames=1):
    """Return the current file name and line number in the program."""
    frame = inspect.currentframe()
    if frame is None:
        msg = "Could not find the current frame"
        raise RuntimeError(msg)
    for num in range(back_frames):
        frame = frame.f_back
        if frame is None:
            msg = f"Could not find the back frame number {num}"
            raise RuntimeError(msg)
    return frame.f_code.co_filename, frame.f_lineno


def load_morphology(path, *, recenter=False):
    """Load a morphology a optionally recenter it."""
    morph = neurom_load_morphology(path)
    if recenter:
        morph = morph.transform(Translation(-morph.soma.center))
    return morph


@disable_loggers("morph_tool.converter")
def save_morphology(
    morph: Morphology,
    morph_path: FileType,
    msg: str | None = None,
    logger: logging.Logger | logging.LoggerAdapter | None = None,
):
    """Export the given morphology to the given path."""
    if msg is None:
        msg = f"Export morphology to {morph_path}"
    logger = sublogger(logger, __name__)
    logger.debug(msg)
    Path(morph_path).parent.mkdir(parents=True, exist_ok=True)

    convert(morph, morph_path, nrn_order=True)
    return morph_path


def create_random_morphologies(
    atlas,
    nb_morphologies: int,
    brain_regions: RegionIdsType | None = None,
    output_morphology_dir: FileType | None = None,
    output_cell_collection: FileType | None = None,
    morphology_prefix: str | None = None,
    rng: SeedType = None,
    logger: logging.Logger | logging.LoggerAdapter | None = None,
):
    """Create some random source points."""
    rng = np.random.default_rng(rng)

    if brain_regions is not None:
        coords, missing_ids = atlas.get_region_points(
            brain_regions, size=nb_morphologies, random_shifts=True, rng=rng
        )
        if missing_ids and logger is not None:
            logger.warning("Could not find the following regions in the atlas: %s", missing_ids)
    else:
        coords, _missing_ids = atlas.get_region_points(
            [-1, 0], size=nb_morphologies, random_shifts=True, inverse=True, rng=rng
        )

    if len(coords) < nb_morphologies and logger is not None:
        logger.error(
            "Not enough voxels found to place source points, found only %s voxels", len(coords)
        )

    dataset = pd.DataFrame(coords, columns=COORDS_COLS).reset_index()
    dataset.index += 1
    dataset.loc[:, "morphology"] = (morphology_prefix or "") + dataset.loc[:, "index"].astype(
        str
    ).str.zfill(len(str(dataset.index.max())))
    dataset.drop(columns=["index"], inplace=True)
    dataset.loc[:, "orientation"] = np.repeat([np.eye(3)], len(dataset), axis=0).tolist()
    dataset.loc[:, "atlas_id"] = atlas.brain_regions.lookup(dataset.loc[:, COORDS_COLS].to_numpy())
    dataset.loc[:, "region"] = dataset.loc[:, "atlas_id"].apply(
        lambda row: atlas.region_map.get(row, attr="acronym", with_ascendants=False)
    )

    if output_morphology_dir is not None:
        output_morphology_dir = Path(output_morphology_dir)
        output_morphology_dir.mkdir(parents=True, exist_ok=True)
        morph_file = output_morphology_dir / dataset.loc[:, "morphology"]
        dataset.loc[:, "morph_file"] = morph_file.astype(str) + ".swc"  # type: ignore[attr-defined]
        for _idx, i in dataset.iterrows():
            morph = MorphIoMorphology()
            morph.soma.points = [i[COORDS_COLS].to_numpy()]
            morph.soma.diameters = [1]
            morph.soma.type = SomaType.SOMA_SINGLE_POINT
            morph.write(i["morph_file"])

    cells = CellCollection.from_dataframe(dataset)

    if output_cell_collection is not None:
        output_cell_collection = Path(output_cell_collection)
        output_cell_collection.parent.mkdir(parents=True, exist_ok=True)
        cells.save(output_cell_collection)

    if logger is not None:
        logger.info("Generated %s random morphologies", len(dataset))

    return cells


def create_dask_dataframe(data: pd.DataFrame, npartitions: int, group_col="morphology"):
    """Ensure all rows of the same group belong to the same partition."""
    ddf = dd.from_pandas(data, npartitions)
    if len(ddf.divisions) > 2:
        groups = np.array_split(data[group_col].unique(), npartitions)
        new_divisions = [
            data.loc[data[group_col].isin(i)].index.min() for i in groups if len(i) > 0
        ] + [data.index.max()]
        ddf = ddf.repartition(divisions=new_divisions)
    return ddf


def parallel_evaluator(
    data,
    func,
    parallel_config: ParallelConfig | None,
    new_columns: list,
    *,
    func_args=None,
    func_kwargs=None,
    startup_func=None,
    shuffle_rows: bool = True,
    progress_bar: bool = False,
    logger: logging.Logger | logging.LoggerAdapter | None = None,
):
    """Create a local cluster and process the given function on the given DataFrame."""
    if logger is None:
        logger = LOGGER
    if parallel_config is None:
        parallel_config = ParallelConfig()
    with disable_distributed_loggers():
        if parallel_config.nb_processes >= 1:
            LOGGER.info("Start parallel computation using %s workers", parallel_config.nb_processes)
            cluster = LocalCluster(n_workers=parallel_config.nb_processes, timeout="60s")
            if startup_func is not None:
                cluster.get_client().run(startup_func)
            parallel_factory = init_parallel_factory(
                "dask_dataframe",
                address=cluster,
            )
        else:
            LOGGER.info("Start computation")
            cluster = None
            shuffle_rows = False
            parallel_factory = init_parallel_factory(None)

        # Extract terminals of each morphology
        results = evaluate(
            data,
            func,
            new_columns,
            parallel_factory=parallel_factory,
            func_args=func_args,
            func_kwargs=func_kwargs,
            shuffle_rows=shuffle_rows,
            progress_bar=progress_bar,
        )

        # Close the Dask cluster if opened
        if cluster is not None:
            parallel_factory.shutdown()
            cluster.close()

    return results


def seed_from_name(name):
    """Build a seed from the name hash."""
    return int(hashlib.sha256(name.encode("ascii")).hexdigest(), 16) % (2**32 - 1)


def get_morph_pts(morph):
    """Return the points of a morphology and handle empty morphologies."""
    try:
        return morph.points
    except ValueError:
        return np.empty((0, 4), dtype=float)


def export_to_tecplot(voxcell_data, filename, title="Default Title"):
    """Export a 3D vector field to a Tecplot binary file."""
    if not has_tecio:
        msg = "The 'tecio' package must be installed to export a vector field into Tecplot format"
        raise RuntimeError(msg)
    shape = voxcell_data.shape
    indices = np.moveaxis(np.mgrid[0 : shape[0], 0 : shape[1], 0 : shape[2]], 0, -1)
    coords = indices * voxcell_data.voxel_dimensions + voxcell_data.offset
    x = coords[:, :, :, 0].ravel().astype(np.float32)
    y = coords[:, :, :, 1].ravel().astype(np.float32)
    z = coords[:, :, :, 2].ravel().astype(np.float32)
    ux = voxcell_data.raw[:, :, :, 0].ravel().astype(np.float32)
    uy = voxcell_data.raw[:, :, :, 1].ravel().astype(np.float32)
    uz = voxcell_data.raw[:, :, :, 2].ravel().astype(np.float32)
    length = np.linalg.norm(voxcell_data.raw, axis=-1).ravel().astype(np.float32)
    tec = {
        "title": title,
        "variables": ["X", "Y", "Z", "Ux", "Uy", "Uz", "length"],
        "zones": [
            {
                "zone_type": ZoneType.ORDERED,
                "ijk": [shape[2], shape[1], shape[0]],
            },
        ],
        "dataset_aux": [],
        "data": [
            {
                "min_max": [
                    [x.min(), x.max()],
                    [y.min(), y.max()],
                    [z.min(), z.max()],
                    [ux.min(), ux.max()],
                    [uy.min(), uy.max()],
                    [uz.min(), uz.max()],
                    [length.min(), length.max()],
                ],
                "data": [
                    x,
                    y,
                    z,
                    ux,
                    uy,
                    uz,
                    length,
                ],
            },
        ],
    }

    with Path(filename).open(mode="wb") as f:
        TecHeader.build_stream(tec, f)
        GreedyRange(gen_zone_struct(7)).build_stream(tec["zones"], f)
        GreedyRange(TecDatasetAux).build_stream(tec.get("dataset_aux", []), f)

        f.write(Float.build(TecplotMarker.EOH))
        for z, d in zip(tec["zones"], tec["data"]):
            gen_data_struct(tec["variables"], z).build_stream(d, f)



"""Compute the long-range trunk properties after clustering."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import logging
from functools import partial

import numpy as np
from neurom import COLS
from neurom import features
from neurom.apps import morph_stats
from neurom.core import Morphology

from axon_synthesis.utils import get_axons

logger = logging.getLogger(__name__)


def angle_between_vectors(p1, p2):
    """Computes the angle in radians between vectors 'p1' and 'p2'.

    Normalizes the input vectors and computes the relative angle
    between them.

        >>> angle_between((1, 0), (0, 1))
        1.5707963267948966
        >>> angle_between((1, 0), (1, 0))
        0.0
        >>> angle_between((1, 0), (-1, 0))
        3.141592653589793
    """
    if p1.shape == p2.shape and np.equal(p1, p2).all():
        return 0.0
    p1 = np.array(p1, ndmin=2)[:, COLS.XYZ]
    p2 = np.array(p2, ndmin=2)[:, COLS.XYZ]
    if p1.shape[0] == 1:
        p1 = np.repeat(p1, p2.shape[0], axis=0)
    if p2.shape[0] == 1:
        p2 = np.repeat(p2, p1.shape[0], axis=0)
    v1 = p1 / np.linalg.norm(p1, axis=1)[:, np.newaxis]
    v2 = p2 / np.linalg.norm(p2, axis=1)[:, np.newaxis]
    dot = np.einsum("ij,ij->i", v1, v2)
    return np.arccos(np.clip(dot, -1.0, 1.0))


def vector(p1: np.ndarray, p2: np.ndarray):
    """Compute vector between two 3D points.

    Args:
        p1: the first point of the vector
        p2: the second point of the vector

    Returns:
        3-vector from p1 - p2
    """
    return np.subtract(p1[..., COLS.XYZ], p2[..., COLS.XYZ])


def section_segment_angles(section, reference=None):
    """Angles between the segments of a section and a reference vector."""
    ref = np.array(reference) if reference is not None else np.array([0, 1, 0])

    seg_vectors = vector(section.points[1:], section.points[:-1])
    return angle_between_vectors(seg_vectors, ref)


@features.feature(shape=(...,), namespace=features.NameSpace.NEURITE)
def segment_angles(neurite, reference=None):
    """Compute the angles between segments of the sections of a neurite."""
    # pylint: disable=protected-access
    func = partial(section_segment_angles, reference=reference)
    return features.neurite._map_segments(func, neurite)  # noqa: SLF001


def compute_trunk_properties(
    trunk_morph: Morphology,
    morph_name: str,
    morph_path: str,
    axon_id: int,
    config_name: str,
    atlas_region_id: int | None,
) -> tuple:
    """Compute the properties of the trunk morphologies listed in the given DataFrame."""
    # Load morph paths
    logger.info("Extracting trunk properties from %s", trunk_morph.name)

    # Compute long-range trunk features that will be used for smoothing and jittering
    long_range_trunk = get_axons(trunk_morph, 0)
    trunk_stats = morph_stats.extract_stats(
        long_range_trunk,
        {
            "neurite": {
                "segment_lengths": {"modes": ["mean", "std"]},
                "segment_meander_angles": {"modes": ["mean", "std"]},
            },
        },
    )["axon"]
    return (
        morph_name,
        morph_path,
        config_name,
        axon_id,
        atlas_region_id,
        trunk_stats["mean_segment_lengths"],
        trunk_stats["std_segment_lengths"],
        trunk_stats["mean_segment_meander_angles"],
        trunk_stats["std_segment_meander_angles"],
    )



"""Cluster the terminal points of a morphology to define a main truk and a set of tufts."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import json
import logging
import os
import pickle
from collections import defaultdict
from copy import deepcopy
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING
from typing import ClassVar

import pandas as pd
from attrs import evolve

from axon_synthesis.atlas import AtlasHelper
from axon_synthesis.base_path_builder import FILE_SELECTION
from axon_synthesis.base_path_builder import BasePathBuilder
from axon_synthesis.inputs.clustering import extract_terminals
from axon_synthesis.inputs.clustering.from_barcodes import (
    compute_clusters as clusters_from_barcodes,
)
from axon_synthesis.inputs.clustering.from_brain_regions import (
    compute_clusters as clusters_from_brain_regions,
)
from axon_synthesis.inputs.clustering.from_sphere_parents import (
    compute_clusters as clusters_from_sphere_parents,
)
from axon_synthesis.inputs.clustering.from_spheres import compute_clusters as clusters_from_spheres
from axon_synthesis.inputs.clustering.plot import plot_cluster_properties
from axon_synthesis.inputs.clustering.plot import plot_clusters
from axon_synthesis.inputs.clustering.utils import DefaultValidatingValidator
from axon_synthesis.inputs.clustering.utils import clusters_basic_tuft
from axon_synthesis.inputs.clustering.utils import compute_shortest_paths
from axon_synthesis.inputs.clustering.utils import create_clustered_morphology
from axon_synthesis.inputs.clustering.utils import export_morph
from axon_synthesis.inputs.clustering.utils import reduce_clusters
from axon_synthesis.inputs.trunk_properties import compute_trunk_properties
from axon_synthesis.typing import FileType
from axon_synthesis.typing import SeedType
from axon_synthesis.typing import Self
from axon_synthesis.utils import COORDS_COLS
from axon_synthesis.utils import MorphNameAdapter
from axon_synthesis.utils import ParallelConfig
from axon_synthesis.utils import get_axons
from axon_synthesis.utils import get_morphology_paths
from axon_synthesis.utils import load_morphology
from axon_synthesis.utils import neurite_to_graph
from axon_synthesis.utils import parallel_evaluator
from axon_synthesis.utils import setup_logger
from axon_synthesis.utils import temp_dir
from axon_synthesis.white_matter_recipe import WhiteMatterRecipe

if TYPE_CHECKING:
    from collections.abc import MutableMapping

    from morphio import PointLevel

LOGGER = logging.getLogger(__name__)

MIN_AXON_POINTS = 5

OUTPUT_COLS = [
    "morphology",
    "morph_file",
    "config_name",
    "axon_id",
    "terminal_id",
    "size",
    *COORDS_COLS,
]

CLUSTERING_FUNCS = {
    "basic": clusters_basic_tuft,
    "sphere": clusters_from_spheres,
    "sphere_parents": clusters_from_sphere_parents,
    "barcode": clusters_from_barcodes,
    "brain_regions": clusters_from_brain_regions,
}


def nodes_to_terminals_mapping(graph, shortest_paths):
    """Map nodes to terminals."""
    node_to_terminals = defaultdict(set)
    for node_id, parent_ids in shortest_paths.items():
        if not graph.nodes[node_id]["is_terminal"]:
            continue
        for j in parent_ids:
            node_to_terminals[j].add(node_id)
    return node_to_terminals


class Clustering(BasePathBuilder):
    """The class to store Clustering data."""

    _filenames: ClassVar[dict] = {
        "CLUSTERING_CONFIGURATIONS_FILENAME": "clustering_configurations.json",
        "FIGURE_DIRNAME": "figures",
        "CLUSTERED_MORPHOLOGIES_DIRNAME": "clustered_morphologies",
        "CLUSTERED_MORPHOLOGIES_PATHS_FILENAME": "clustered_morphologies_paths.csv",
        "CLUSTERED_TERMINALS_FILENAME": "clustered_terminals.csv",
        "TERMINALS_DIRNAME": "clustered_terminals",
        "TRUNK_MORPHOLOGIES_DIRNAME": "trunk_morphologies",
        "TRUNK_MORPHOLOGIES_PATHS_FILENAME": "trunk_morphologies_paths.csv",
        "TRUNK_PROPS_FILENAME": "trunk_properties.json",
        "TUFT_MORPHOLOGIES_DIRNAME": "tuft_morphologies",
        "TUFT_MORPHOLOGIES_PATHS_FILENAME": "tuft_morphologies_paths.csv",
        "TUFT_PROPS_FILENAME": "tuft_properties.json",
        "TUFT_PROPS_PLOT_FILENAME": "tuft_properties.pdf",
    }
    _optional_keys: ClassVar[set[str]] = {
        "CLUSTERED_TERMINALS_FILENAME",
        "CLUSTERED_MORPHOLOGIES_DIRNAME",
        "CLUSTERED_MORPHOLOGIES_PATHS_FILENAME",
        "FIGURE_DIRNAME",
        "TERMINALS_DIRNAME",
        "TRUNK_MORPHOLOGIES_DIRNAME",
        "TRUNK_MORPHOLOGIES_PATHS_FILENAME",
        "TUFT_MORPHOLOGIES_DIRNAME",
        "TUFT_MORPHOLOGIES_PATHS_FILENAME",
        "TUFT_PROPS_PLOT_FILENAME",
    }

    PARAM_SCHEMA: ClassVar[dict] = {
        "type": "object",
        "patternProperties": {
            ".*": {
                "oneOf": [
                    {
                        # For 'sphere' clustering mode
                        "additionalProperties": False,
                        "properties": {
                            "method": {
                                "type": "string",
                                "enum": ["sphere"],
                            },
                            "sphere_radius": {
                                "type": "number",
                                "exclusiveMinimum": 0,
                                "default": 100,
                            },
                            "min_size": {
                                "type": "integer",
                                "minimum": 1,
                                "default": 10,
                            },
                        },
                        "required": [
                            "method",
                        ],
                        "type": "object",
                    },
                    {
                        # For 'sphere_parents' clustering mode
                        "additionalProperties": False,
                        "properties": {
                            "method": {
                                "type": "string",
                                "enum": ["sphere_parents"],
                            },
                            "sphere_radius": {
                                "type": "number",
                                "exclusiveMinimum": 0,
                                "default": 100,
                            },
                            "max_path_distance": {
                                "type": "number",
                                "exclusiveMinimum": 0,
                            },
                        },
                        "required": [
                            "method",
                        ],
                        "type": "object",
                    },
                    {
                        # For 'brain_regions' clustering mode
                        "additionalProperties": False,
                        "properties": {
                            "method": {
                                "type": "string",
                                "enum": ["brain_regions"],
                            },
                            "wm_unnesting": {
                                "type": "boolean",
                                "default": True,
                            },
                        },
                        "required": [
                            "method",
                        ],
                        "type": "object",
                    },
                    {
                        # For 'barcode' clustering mode
                        "additionalProperties": False,
                        "properties": {
                            "method": {
                                "type": "string",
                                "enum": ["barcode"],
                            },
                        },
                        "required": [
                            "method",
                        ],
                        "type": "object",
                    },
                ],
            },
        },
    }
    PARAM_SCHEMA_VALIDATOR = DefaultValidatingValidator(PARAM_SCHEMA)  # type: ignore[operator]

    def __init__(self, path: FileType, parameters: dict, **kwargs):
        """The Clustering constructor.

        Args:
            path: The base path used to build the relative paths.
            parameters: The parameters used for clustering.
            **kwargs: The keyword arguments are passed to the base constructor.
        """
        super().__init__(path, **kwargs)

        if kwargs.get("create", False):
            self.create_tree()

        # Process parameters
        parameters = deepcopy(parameters)
        self.PARAM_SCHEMA_VALIDATOR.validate(parameters)
        self._parameters = parameters

        # Clustering results
        self.clustered_terminals: pd.DataFrame | None = None
        self.clustered_morph_paths: pd.DataFrame | None = None
        self.trunk_properties: pd.DataFrame | None = None
        self.trunk_morph_paths: pd.DataFrame | None = None
        self.tuft_properties: pd.DataFrame | None = None
        self.tuft_morph_paths: pd.DataFrame | None = None

    @property
    def parameters(self):
        """Return the parameters used for clustering."""
        return self._parameters

    def create_tree(self):
        """Create the associated directories."""
        self.path.mkdir(parents=True, exist_ok=True)
        for k, v in self:
            if k.endswith("_DIRNAME"):
                v.mkdir(parents=True, exist_ok=True)

    def plot_cluster_properties(self):
        """Plot cluster properties."""
        if self.tuft_properties is not None:
            plot_cluster_properties(self.tuft_properties, self.TUFT_PROPS_PLOT_FILENAME)
            LOGGER.info(
                "Exported figure of cluster properties to %s",
                self.TUFT_PROPS_PLOT_FILENAME,
            )
        else:
            LOGGER.warning(
                "Can't export figure of cluster properties because they were not computed yet"
            )

    def save(self):
        """Save the clustering data to the associated path."""
        # Export long-range trunk properties
        if self.trunk_properties is not None:
            with self.TRUNK_PROPS_FILENAME.open(mode="w", encoding="utf-8") as f:
                json.dump(self.trunk_properties.to_dict("records"), f, indent=4)
            LOGGER.info("Exported trunk properties to %s", self.TRUNK_PROPS_FILENAME)

        # Export tuft properties
        if self.tuft_properties is not None:
            with self.TUFT_PROPS_FILENAME.open(mode="w", encoding="utf-8") as f:
                json.dump(self.tuft_properties.to_dict("records"), f, indent=4)
            LOGGER.info("Exported tuft properties to %s", self.TUFT_PROPS_FILENAME)

        # Export the terminals
        if self.clustered_terminals is not None:
            self.clustered_terminals.to_csv(self.CLUSTERED_TERMINALS_FILENAME, index=False)
            LOGGER.info("Exported cluster terminals to %s", self.CLUSTERED_TERMINALS_FILENAME)

        # Export morphology paths
        if self.clustered_morph_paths is not None:
            self.clustered_morph_paths.to_csv(
                self.CLUSTERED_MORPHOLOGIES_PATHS_FILENAME, index=False
            )
            LOGGER.info(
                "Exported clustered morphologies paths to %s",
                self.CLUSTERED_MORPHOLOGIES_PATHS_FILENAME,
            )

        # Export trunk morphology paths
        if self.trunk_morph_paths is not None:
            self.trunk_morph_paths.to_csv(self.TRUNK_MORPHOLOGIES_PATHS_FILENAME, index=False)
            LOGGER.info(
                "Exported trunk morphologies paths to %s",
                self.TRUNK_MORPHOLOGIES_PATHS_FILENAME,
            )

        # Export trunk morphology paths
        if self.tuft_morph_paths is not None:
            self.tuft_morph_paths.to_csv(self.TUFT_MORPHOLOGIES_PATHS_FILENAME, index=False)
            LOGGER.info(
                "Exported tuft morphologies paths to %s",
                self.TUFT_MORPHOLOGIES_PATHS_FILENAME,
            )

        # Export the clustering configurations so they can can be retrieved afterwards
        with self.CLUSTERING_CONFIGURATIONS_FILENAME.open(mode="w", encoding="utf-8") as f:
            json.dump(self.parameters, f, indent=4)
        LOGGER.info("Exported clustering parameters to %s", self.CLUSTERING_CONFIGURATIONS_FILENAME)

    @classmethod
    def load(
        cls,
        path: FileType,
        file_selection: FILE_SELECTION = FILE_SELECTION.NONE,
        *,
        allow_missing: bool = False,
        **kwargs,
    ) -> Self:
        """Load the clustering data from a given directory."""
        path = Path(path)
        paths = cls.build_default_paths(path)

        # Import the clustering configurations
        with paths["CLUSTERING_CONFIGURATIONS_FILENAME"].open(encoding="utf-8") as f:
            parameters = json.load(f)

        # Create the object
        obj = cls(path, parameters)

        obj.update_from_dict("trunk_properties_file", "TRUNK_PROPS_FILENAME", kwargs)
        obj.update_from_dict("tuft_properties_file", "TUFT_PROPS_FILENAME", kwargs)

        # Load data if they exist
        if file_selection <= FILE_SELECTION.REQUIRED_ONLY:
            try:
                obj.assert_exists(file_selection=FILE_SELECTION.REQUIRED_ONLY)
            except FileNotFoundError:
                if not allow_missing:
                    raise
            else:
                with obj.TRUNK_PROPS_FILENAME.open(encoding="utf-8") as f:
                    obj.trunk_properties = pd.read_json(
                        obj.TRUNK_PROPS_FILENAME, dtype={"morphology": str, "population_id": str}
                    )
                with obj.TUFT_PROPS_FILENAME.open(encoding="utf-8") as f:
                    obj.tuft_properties = pd.read_json(
                        f, dtype={"morphology": str, "population_id": str}
                    )
        if file_selection <= FILE_SELECTION.ALL or file_selection == FILE_SELECTION.OPTIONAL_ONLY:
            try:
                obj.assert_exists(file_selection=FILE_SELECTION.OPTIONAL_ONLY)
            except FileNotFoundError:
                if not allow_missing:
                    raise
            else:
                obj.clustered_terminals = pd.read_csv(
                    obj.CLUSTERED_TERMINALS_FILENAME, dtype={"morphology": str}
                )
                obj.clustered_morph_paths = pd.read_csv(obj.CLUSTERED_MORPHOLOGIES_PATHS_FILENAME)
                obj.trunk_morph_paths = pd.read_csv(obj.TRUNK_MORPHOLOGIES_PATHS_FILENAME)
                obj.tuft_morph_paths = pd.read_csv(obj.TUFT_MORPHOLOGIES_PATHS_FILENAME)

        return obj


def extract_morph_name_from_filename(df, file_col="morph_file", name_col="morphology"):
    """Add a 'morphology' column in the given DataFrame computed from the 'morph_file' column."""
    df[name_col] = df[file_col].apply(lambda x: Path(x).stem)
    return df


def export_clusters(
    clustering, trunk_props, cluster_props, all_terminal_points, morph_paths, *, debug=False
):
    """Export cluster data."""
    # Export long-range trunk properties
    clustering.trunk_properties = pd.DataFrame(
        trunk_props,
        columns=[
            "morphology",
            "morph_file",
            "config_name",
            "axon_id",
            "atlas_region_id",
            "mean_segment_lengths",
            "std_segment_lengths",
            "mean_segment_meander_angles",
            "std_segment_meander_angles",
        ],
    ).sort_values(["morphology", "config_name", "axon_id"])

    # Export tuft properties
    clustering.tuft_properties = pd.DataFrame(
        cluster_props,
        columns=[
            "morphology",
            "morph_file",
            "config_name",
            "axon_id",
            "tuft_id",
            "center_coords",
            "common_ancestor_id",
            "common_ancestor_x",
            "common_ancestor_y",
            "common_ancestor_z",
            "path_distance",
            "radial_distance",
            "path_length",
            "max_path_extent",
            "size",
            "orientation",
            "mean_tuft_length",
            "atlas_region_id",
            "population_id",
            "barcode",
        ],
    ).sort_values(["morphology", "config_name", "axon_id"])

    # Plot cluster properties
    if debug:
        clustering.plot_cluster_properties()

    # Export the terminals
    clustering.clustered_terminals = (
        pd.DataFrame(all_terminal_points, columns=OUTPUT_COLS)
        .fillna({"size": 1})
        .astype({"size": int})
        .sort_values(["morphology", "config_name", "axon_id", "terminal_id"])
    )
    # TODO: Check if some terminals are missing here and if can remove the merge below
    # clustering.clustered_terminals = pd.merge(
    #     clustering.clustered_terminals,
    #     terminals.groupby(
    #         ["morph_file", "axon_id", "tuft_id", "config"]
    #     ).size().rename("size"),
    #     left_on=["morph_file", "axon_id", "terminal_id"],
    #     right_on=["morph_file", "axon_id", "tuft_id"],
    #     how="left",
    # )

    # Export morphology paths
    clustering.clustered_morph_paths = pd.DataFrame(
        morph_paths["clustered"],
        columns=["morphology", "morph_file", "config_name", "axon_id", "morph_path"],
    ).sort_values(["morphology", "config_name", "axon_id"])

    clustering.trunk_morph_paths = pd.DataFrame(
        morph_paths["trunks"],
        columns=["morphology", "morph_file", "config_name", "axon_id", "morph_path"],
    ).sort_values(["morphology", "config_name", "axon_id"])

    if debug:
        clustering.tuft_morph_paths = pd.DataFrame(
            morph_paths["tufts"],
            columns=["morphology", "morph_file", "config_name", "axon_id", "tuft_id", "morph_path"],
        ).sort_values(["morphology", "config_name", "axon_id", "tuft_id"])

    # TODO: Should the 'use_ancestors' mode be moved here from the graph creation step?


def cluster_one_morph(
    morph_path: FileType,
    clustering: Clustering,
    atlas: AtlasHelper | None,
    wmr: WhiteMatterRecipe | None,
    projection_pop_numbers: pd.DataFrame | None,
    bouton_density: float | None,
    *,
    morph_name: str | None = None,
    debug: bool = False,
    rng: SeedType = None,
    parallel_config: ParallelConfig | None = None,
):
    """Run clustering on one morphology."""
    parallel_config = ParallelConfig() if parallel_config is None else parallel_config

    morph_name = Path(morph_path).name if morph_name is None else morph_name
    morph_path = str(morph_path)

    morph_custom_logger = MorphNameAdapter(LOGGER, extra={"morph_name": morph_name})

    pts = extract_terminals.process_morph(morph_path, morph_name)
    if len(pts) == 0:
        morph_custom_logger.warning("The morphology has no axon point")
        return ClusteringResult([], [], [], {})
    terminals = pd.DataFrame(
        pts,
        columns=["morphology", "morph_file", "axon_id", "terminal_id", "section_id", *COORDS_COLS],
    )
    terminals[["config", "tuft_id"]] = None, -1

    brain_regions = atlas.brain_regions if atlas is not None else None

    all_terminal_points: list[tuple] = []
    cluster_props: list[tuple] = []
    trunk_props: list[tuple] = []
    morph_paths: MutableMapping[str, list] = defaultdict(list)

    morph_custom_logger.debug("%s points", len(terminals))

    # Load the morphology
    morph = load_morphology(morph_path)

    # Get the source brain region
    atlas_region_id = brain_regions.lookup(morph.soma.center) if brain_regions is not None else None

    # Run the clustering function on each axon
    for axon_id, axon in enumerate(get_axons(morph)):
        # Create a graph for each axon and compute the shortest paths from the soma to all
        # terminals
        nodes, edges, directed_graph = neurite_to_graph(axon)
        shortest_paths = compute_shortest_paths(directed_graph)
        node_to_terminals = nodes_to_terminals_mapping(directed_graph, shortest_paths)
        axon_custom_logger = MorphNameAdapter(
            LOGGER, extra={"morph_name": morph_name, "axon_id": axon_id}
        )

        for config_name, config in clustering.parameters.items():
            clustering_method = config["method"]
            if clustering_method == "brain_regions" and (atlas is None or wmr is None):
                msg = (
                    "The atlas and wmr can not be None when the clustering method is "
                    "'brain_regions'."
                )
                raise ValueError(msg)
            if len(axon.points) < MIN_AXON_POINTS:
                axon_custom_logger.warning(
                    "The axon %s of %s is clustered with basic algorithm because it has only "
                    "%s points while at least 5 points are needed for other clustering "
                    "algorithms",
                    axon_id,
                    morph_name,
                    len(axon.points),
                )
                clustering_method = "basic"
            axon_group = terminals.loc[terminals["axon_id"] == axon_id].copy(deep=True)
            axon_group["config_name"] = config_name
            axon_group = axon_group.merge(
                nodes.reset_index().rename(columns={"id": "graph_node_id"})[
                    ["section_id", "graph_node_id"]
                ],
                on="section_id",
                how="left",
            )
            suffix = f"_{config_name}_{axon_id}"
            clustering_kwargs = {
                "atlas": atlas,
                "wmr": wmr,
                "config": config,
                "config_name": config_name,
                "morph": morph,
                "axon": axon,
                "axon_id": axon_id,
                "group_name": morph_name,
                "group_path": morph_path,
                "group": axon_group,
                "nodes": nodes,
                "edges": edges,
                "directed_graph": directed_graph,
                "shortest_paths": shortest_paths,
                "node_to_terminals": node_to_terminals,
                "output_cols": OUTPUT_COLS,
                "clustered_morphologies_path": clustering.CLUSTERED_MORPHOLOGIES_DIRNAME,
                "trunk_morphologies_path": clustering.TRUNK_MORPHOLOGIES_DIRNAME,
                "tuft_morphologies_path": clustering.TUFT_MORPHOLOGIES_DIRNAME,
                "figure_path": clustering.FIGURE_DIRNAME,
                "nb_workers": parallel_config.nb_processes,
                "debug": debug,
                "logger": axon_custom_logger,
            }
            new_terminal_points, tuft_ids = CLUSTERING_FUNCS[clustering_method](  # type: ignore[operator]
                **clustering_kwargs
            )

            # Add the cluster to the final points
            all_terminal_points.extend(new_terminal_points)

            # Propagate cluster IDs
            axon_group["tuft_id"] = tuft_ids

            axon_custom_logger.info(
                "%s (axon %s): %s points after merge",
                morph_name,
                axon_id,
                len(new_terminal_points),
            )

            cluster_df = pd.DataFrame(new_terminal_points, columns=OUTPUT_COLS)

            # Reduce clusters to one section
            sections_to_add: MutableMapping[int, PointLevel] = defaultdict(list)
            kept_path = reduce_clusters(
                axon_group,
                morph_path,
                morph_name,
                morph,
                axon,
                axon_id,
                cluster_df,
                directed_graph,
                nodes,
                sections_to_add,
                morph_paths,
                cluster_props,
                shortest_paths,
                bouton_density,
                brain_regions,
                atlas_region_id,
                atlas.orientations if atlas is not None else None,
                projection_pop_numbers=projection_pop_numbers,
                export_tuft_morph_dir=clustering.TUFT_MORPHOLOGIES_DIRNAME if debug else None,
                config_name=config_name,
                rng=rng,
                logger=axon_custom_logger,
            )

            # Create the clustered morphology
            clustered_morph, trunk_morph = create_clustered_morphology(
                morph,
                morph_name,
                axon_id,
                kept_path,
                sections_to_add,
                suffix=suffix,
            )

            # Compute trunk properties
            trunk_props.append(
                compute_trunk_properties(
                    trunk_morph,
                    str(morph_name),
                    str(morph_path),
                    axon_id,
                    config_name,
                    atlas_region_id,
                )
            )

            # Export the trunk and clustered morphologies
            morph_paths["clustered"].append(
                (
                    morph_name,
                    morph_path,
                    config_name,
                    axon_id,
                    export_morph(
                        clustering.CLUSTERED_MORPHOLOGIES_DIRNAME,
                        morph_name,
                        clustered_morph,
                        "clustered",
                        suffix=suffix,
                    ),
                ),
            )
            morph_paths["trunks"].append(
                (
                    morph_name,
                    morph_path,
                    config_name,
                    axon_id,
                    export_morph(
                        clustering.TRUNK_MORPHOLOGIES_DIRNAME,
                        morph_name,
                        trunk_morph,
                        "trunk",
                        suffix=suffix,
                    ),
                ),
            )

            # Plot the clusters
            if debug:
                plot_clusters(
                    morph,
                    clustered_morph,
                    axon_group,
                    morph_name,
                    cluster_df,
                    clustering.FIGURE_DIRNAME
                    / f"{Path(str(morph_name)).with_suffix('').name}{suffix}.html",
                )
                axon_group.to_csv(
                    clustering.TERMINALS_DIRNAME
                    / f"terminals_{Path(str(morph_name)).with_suffix('').name}{suffix}.csv"
                )
                cluster_df.to_csv(
                    clustering.TERMINALS_DIRNAME
                    / f"clusters_{Path(str(morph_name)).with_suffix('').name}{suffix}.csv"
                )

    return ClusteringResult(trunk_props, cluster_props, all_terminal_points, morph_paths)


def _wrapper(data: dict, **kwargs: dict) -> dict:
    """Wrap process_morph() for parallel computation."""
    all_kwargs = {**data, **kwargs}
    trunk_props, cluster_props, all_terminal_points, morph_paths = cluster_one_morph(**all_kwargs)

    return {
        "trunk_props": trunk_props,
        "cluster_props": cluster_props,
        "terminal_points": all_terminal_points,
        "morph_paths": morph_paths,
    }


class ClusteringResult:
    """Class to store clustering result for one morphology."""

    TRUNK_PROPS_FILENAME = "trunk_props"
    CLUSTER_PROPS_FILENAME = "cluster_props"
    ALL_TERMINAL_POINTS_FILENAME = "all_terminal_points"
    MORPH_PATH_FILENAME = "morph_paths"

    def __init__(self, trunk_props, cluster_props, all_terminal_points, morph_paths):
        """The ClusteringResult constructor."""
        self.trunk_props = trunk_props
        self.cluster_props = cluster_props
        self.all_terminal_points = all_terminal_points
        self.morph_paths = morph_paths

    def __getitem__(self, name):
        """Behave like a dict."""
        return getattr(self, name)

    def __iter__(self):
        """Make the class iterable."""
        yield self.trunk_props
        yield self.cluster_props
        yield self.all_terminal_points
        yield self.morph_paths

    def save(self, **kwargs):
        """Save internals to a temporary directory."""
        tmpdir = temp_dir(**kwargs)
        dir_path = Path(tmpdir.name)
        with (dir_path / self.TRUNK_PROPS_FILENAME).open(mode="wb") as f:
            pickle.dump(self.trunk_props, f)
        with (dir_path / self.CLUSTER_PROPS_FILENAME).open(mode="wb") as f:
            pickle.dump(self.cluster_props, f)
        with (dir_path / self.ALL_TERMINAL_POINTS_FILENAME).open(mode="wb") as f:
            pickle.dump(self.all_terminal_points, f)
        with (dir_path / self.MORPH_PATH_FILENAME).open(mode="wb") as f:
            pickle.dump(self.morph_paths, f)
        return tmpdir

    @classmethod
    def load(cls, path: FileType) -> Self:
        """Load internals from a directory."""
        dir_path = Path(path)
        with (dir_path / cls.TRUNK_PROPS_FILENAME).open(mode="rb") as f:
            trunk_props = pickle.load(f)  # noqa: S301
        with (dir_path / cls.CLUSTER_PROPS_FILENAME).open(mode="rb") as f:
            cluster_props = pickle.load(f)  # noqa: S301
        with (dir_path / cls.ALL_TERMINAL_POINTS_FILENAME).open(mode="rb") as f:
            all_terminal_points = pickle.load(f)  # noqa: S301
        with (dir_path / cls.MORPH_PATH_FILENAME).open(mode="rb") as f:
            morph_paths = pickle.load(f)  # noqa: S301
        return cls(trunk_props, cluster_props, all_terminal_points, morph_paths)


# This might be useful to optimize Dask communication
# @dask_serialize.register(ClusteringResult)
# def serialize(res: ClusteringResult) -> tuple[dict, list[bytes]]:
#     header = {}
#     tmpdir = res.save()
#     LOGGER.critical("Saved to %s", tmpdir.name)
#     frames = [str(tmpdir.name).encode()]
#     return header, frames

# @dask_deserialize.register(ClusteringResult)
# def deserialize(header: dict, frames: list[bytes]) -> ClusteringResult:
#     LOGGER.critical("Load from %s", frames[0].decode())
#     return ClusteringResult.load(frames[0].decode())


def cluster_morphologies(
    morph_dir: FileType,
    clustering_parameters: dict,
    output_path: FileType,
    *,
    atlas: AtlasHelper | None,
    wmr: WhiteMatterRecipe | None,
    pop_neuron_numbers: pd.DataFrame | None,
    bouton_density: float | None,
    debug: bool = False,
    rng: SeedType = None,
    parallel_config: ParallelConfig | None = None,
) -> Clustering:
    """Compute the cluster of all morphologies of the given directory."""
    if parallel_config is None:
        parallel_config = ParallelConfig()

    clustering = Clustering(output_path, clustering_parameters, create=True)

    if clustering.path.exists():
        LOGGER.warning(
            "The '%s' folder already exists, the new morphologies will be added to it",
            clustering.path,
        )

    LOGGER.info(
        "Clustering morphologies using the following configuration: %s",
        clustering.parameters,
    )

    projection_pop_numbers = (
        wmr.projection_targets.merge(
            pop_neuron_numbers[
                ["pop_raw_name", "atlas_region_volume", "pop_neuron_numbers"]
            ].drop_duplicates(),
            left_on="target_population_name",
            right_on="pop_raw_name",
            how="left",
            suffixes=("", "_target"),
        )
        if wmr is not None and wmr.projection_targets is not None and pop_neuron_numbers is not None
        else None
    )

    morphologies = get_morphology_paths(morph_dir)

    if len(morphologies) == 0:
        LOGGER.error("No morphology file found in '%s'", morph_dir)
        return clustering
    LOGGER.info("Found %s morphology files in '%s'", len(morphologies), morph_dir)
    parallel_config_tmp = evolve(
        parallel_config,
        nb_processes=min(len(morphologies), parallel_config.nb_processes, os.cpu_count() or 0),
    )

    # TODO: Should use map_partitions and load the atlas and other data from the workers
    results = parallel_evaluator(
        morphologies,
        _wrapper,
        parallel_config_tmp,
        [
            ["trunk_props", None],
            ["cluster_props", None],
            ["terminal_points", None],
            ["morph_paths", None],
        ],
        func_kwargs={
            "clustering": clustering,
            "atlas": atlas,
            "wmr": wmr,
            "projection_pop_numbers": projection_pop_numbers,
            "bouton_density": bouton_density,
            "debug": debug,
            "rng": rng,
            "parallel_config": parallel_config_tmp,
        },
        progress_bar=False,
        startup_func=partial(setup_logger, level=logging.getLevelName(LOGGER.getEffectiveLevel())),
    )

    trunk_props = results["trunk_props"].dropna().explode().dropna().tolist()
    cluster_props = results["cluster_props"].dropna().explode().dropna().tolist()
    all_terminal_points = results["terminal_points"].dropna().explode().dropna().tolist()
    paths = results["morph_paths"].dropna().apply(pd.Series).dropna()
    morph_paths = {col: paths[col].explode().tolist() for col in paths.columns}

    export_clusters(
        clustering, trunk_props, cluster_props, all_terminal_points, morph_paths, debug=debug
    )

    return clustering



"""Tests suite for the axon-synthesis package."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

from contextlib import contextmanager
from pathlib import Path

import matplotlib as mpl

TEST_ROOT = Path(__file__).parent
EXAMPLES = TEST_ROOT.parent / "examples"
DATA = TEST_ROOT / "data"


class SharedData:
    """A class used to pass data from a test to another."""

    create_inputs_path = None


@contextmanager
def use_matplotlib_backend(new_backend):
    """A context manager to set a new temporary backend to matplotlib then restore the old one.

    Args:
        new_backend (str): The name of the backend to use in this context.
    """
    old_backend = mpl.get_backend()
    mpl.use(new_backend)
    try:
        yield
    finally:
        mpl.use(old_backend)



"""Entries of the Command Line Interface dedicated to synthesis."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
# LICENSE HEADER MANAGED BY add-license-header

import functools
from collections.abc import Callable

import click
from attrs import fields_dict
from click_option_group import optgroup

from axon_synthesis.cli.common import atlas_kwargs_to_config
from axon_synthesis.cli.common import atlas_options
from axon_synthesis.cli.common import parallel_kwargs_to_config
from axon_synthesis.cli.common import parallel_options
from axon_synthesis.cli.utils import GlobalConfig
from axon_synthesis.cli.utils import ListParam
from axon_synthesis.synthesis import SynthesisConfig
from axon_synthesis.synthesis import synthesize_axons
from axon_synthesis.synthesis.main_trunk.create_graph import CreateGraphConfig
from axon_synthesis.synthesis.main_trunk.post_process import PostProcessConfig
from axon_synthesis.synthesis.outputs import OutputConfig


def create_graph_options(func):
    """Decorate a click command to add Atlas-specific options."""

    @optgroup.group(
        "Create Graph parameters",
        help="Parameters used to create the graph on which the Steiner Tree is computed",
    )
    @optgroup.option(
        "--create-graph-intermediate-number",
        type=click.IntRange(min=0),
        help="The number of intermediate points added before Voronoï process",
    )
    @optgroup.option(
        "--create-graph-min-intermediate-distance",
        type=click.FloatRange(min=0, min_open=True),
        help="The min distance between two successive intermediate points",
    )
    @optgroup.option(
        "--create-graph-min-random-point-distance",
        type=click.FloatRange(min=0, min_open=True),
        help="The min distance used to add random points",
    )
    @optgroup.option(
        "--create-graph-random-max-tries",
        type=click.IntRange(min=1),
        help="The max number of tries to place a new random point",
    )
    @optgroup.option(
        "--create-graph-random-point-bbox-buffer",
        type=click.FloatRange(min=0),
        help="The distance used to add a buffer around the bbox of the points",
    )
    @optgroup.option(
        "--create-graph-voronoi-steps",
        type=click.IntRange(min=1),
        help="The number of Voronoi steps",
    )
    @optgroup.option(
        "--create-graph-duplicate_precision",
        type=click.FloatRange(min=0, min_open=True),
        help="The precision used to detect duplicated points",
    )
    @optgroup.option(
        "--create-graph-use-orientation-penalty/--create-graph-no-use-orientation-penalty",
        default=None,
        help="If set to True, a penalty is added to edges whose direction is not radial",
    )
    @optgroup.option(
        "--create-graph-orientation-penalty-exponent",
        type=click.FloatRange(min=0),
        help="The exponent used for the orientation penalty",
    )
    @optgroup.option(
        "--create-graph-orientation-penalty-amplitude",
        type=click.FloatRange(min=0, min_open=True),
        help="The amplitude used for the orientation penalty",
    )
    @optgroup.option(
        "--create-graph-use-depth-penalty/--create-graph-no-use-depth-penalty",
        default=None,
        help=(
            "If set to True, a penalty is added to edges whose direction is not parallel to the "
            "iso-depth curves"
        ),
    )
    @optgroup.option(
        "--create-graph-depth-penalty-sigma",
        type=click.FloatRange(min=0, min_open=True),
        help="The sigma used for depth penalty",
    )
    @optgroup.option(
        "--create-graph-depth-penalty-amplitude",
        type=click.FloatRange(min=0, min_open=True),
        help="The amplitude of the depth penalty",
    )
    @optgroup.option(
        "--create-graph-preferred-regions",
        type=ListParam(),
        help="The list of brain regions in which edge weights are divided by the preferring factor",
    )
    @optgroup.option(
        "--create-graph-preferred-region-min-random-point-distance",
        type=click.FloatRange(min=0, min_open=True),
        help="The min distance used to pick random points in preferred regions",
    )
    @optgroup.option(
        "--create-graph-preferring-sigma",
        type=click.FloatRange(min=0, min_open=True),
        help="The sigma used to compute the preferring factor for the given regions",
    )
    @optgroup.option(
        "--create-graph-preferring-amplitude",
        type=click.FloatRange(min=0, min_open=True),
        help="The amplitude used to compute the preferring factor for the given regions",
    )
    @optgroup.option(
        "--create-graph-use-terminal-penalty/--create-graph-no-use-terminal-penalty",
        default=None,
        help="If enabled, a penalty is added to edges that are connected to a",
    )
    @functools.wraps(func)
    def wrapper_create_graph_options(*args, **kwargs) -> Callable:
        return func(*args, **kwargs)

    return wrapper_create_graph_options


def create_graph_kwargs_to_config(config) -> None:
    """Extract the atlas arguments from given config to create an AtlasConfig object."""
    kwargs = {
        "intermediate_number": config.pop("create_graph_intermediate_number", None),
        "min_intermediate_distance": config.pop("create_graph_min_intermediate_distance", None),
        "min_random_point_distance": config.pop("create_graph_min_random_point_distance", None),
        "random_max_tries": config.pop("create_graph_random_max_tries", None),
        "random_point_bbox_buffer": config.pop("create_graph_random_point_bbox_buffer", None),
        "voronoi_steps": config.pop("create_graph_voronoi_steps", None),
        "duplicate_precision": config.pop("create_graph_duplicate_precision", None),
        "orientation_penalty_exponent": config.pop(
            "create_graph_orientation_penalty_exponent", None
        ),
        "orientation_penalty_amplitude": config.pop(
            "create_graph_orientation_penalty_amplitude", None
        ),
        "depth_penalty_sigma": config.pop("create_graph_depth_penalty_sigma", None),
        "depth_penalty_amplitude": config.pop("create_graph_depth_penalty_amplitude", None),
        "preferred_regions": config.pop("create_graph_preferred_regions", None),
        "preferred_region_min_random_point_distance": config.pop(
            "create_graph_preferred_region_min_random_point_distance", None
        ),
        "preferring_sigma": config.pop("create_graph_preferring_sigma", None),
        "preferring_amplitude": config.pop("create_graph_preferring_amplitude", None),
        "use_depth_penalty": config.pop("create_graph_use_depth_penalty", None),
        "use_orientation_penalty": config.pop("create_graph_use_orientation_penalty", None),
        "use_terminal_penalty": config.pop("create_graph_use_terminal_penalty", None),
    }
    kwargs = {k: v for k, v in kwargs.items() if v is not None}

    config["create_graph_config"] = CreateGraphConfig(**kwargs)


def post_process_options(func):
    """Decorate a click command to add Atlas-specific options."""

    @optgroup.group(
        "Long-range trunk Post-Processing parameters",
        help="Parameters used to post-process the long-range trunk of the morphology",
    )
    @optgroup.option(
        "--post-processing-disable/--post-processing-enable",
        default=None,
        help="If disabled, the long-range trunk is not post-processed with random walk",
    )
    @optgroup.option(
        "--post-processing-history-path-length",
        type=click.FloatRange(min=0),
        help="The length used to compute the random walk history",
    )
    @optgroup.option(
        "--post-processing-default-history-path-length",
        type=click.FloatRange(min=0, min_open=True),
        help="The coefficient used to compute the history path length when it is not provided",
    )
    @optgroup.option(
        "--post-processing-global-target-coeff",
        type=click.FloatRange(min=0),
        help="The coefficient applied to the global target term",
    )
    @optgroup.option(
        "--post-processing-global-target-sigma-coeff",
        type=click.FloatRange(min=0, min_open=True),
        help="The sigma coefficient applied to the global target term",
    )
    @optgroup.option(
        "--post-processing-target-coeff",
        type=click.FloatRange(min=0),
        help="The coefficient applied to the next target term",
    )
    @optgroup.option(
        "--post-processing-target-sigma-coeff",
        type=click.FloatRange(min=0, min_open=True),
        help="The sigma coefficient applied to the next target term",
    )
    @optgroup.option(
        "--post-processing-random-coeff",
        type=click.FloatRange(min=0),
        help="The coefficient applied to the random term",
    )
    @optgroup.option(
        "--post-processing-history-coeff",
        type=click.FloatRange(min=0),
        help="The coefficient applied to the history term",
    )
    @optgroup.option(
        "--post-processing-history-sigma-coeff",
        type=click.FloatRange(min=0, min_open=True),
        help="The sigma coefficient applied to the history term",
    )
    @optgroup.option(
        "--post-processing-length-coeff",
        type=click.FloatRange(min=0, min_open=True),
        help="The coefficient applied to step length",
    )
    @optgroup.option(
        "--post-processing-max-random-direction-picks",
        type=click.IntRange(min=1),
        help="The maximum number of random direction picks",
    )
    @functools.wraps(func)
    def wrapper_post_processing_options(*args, **kwargs) -> Callable:
        return func(*args, **kwargs)

    return wrapper_post_processing_options


def post_process_kwargs_to_config(config) -> None:
    """Extract the post-process arguments from given config to create a PostProcessConfig object."""
    kwargs = {
        "history_path_length": config.pop("post_processing_history_path_length", None),
        "default_history_path_length": config.pop(
            "post_processing_default_history_path_length", None
        ),
        "global_target_coeff": config.pop("post_processing_global_target_coeff", None),
        "global_target_sigma_coeff": config.pop("post_processing_global_target_sigma_coeff", None),
        "target_coeff": config.pop("post_processing_target_coeff", None),
        "target_sigma_coeff": config.pop("post_processing_target_sigma_coeff", None),
        "random_coeff": config.pop("post_processing_random_coeff", None),
        "history_coeff": config.pop("post_processing_history_coeff", None),
        "history_sigma_coeff": config.pop("post_processing_history_sigma_coeff", None),
        "length_coeff": config.pop("post_processing_length_coeff", None),
        "max_random_direction_picks": config.pop(
            "post_processing_max_random_direction_picks", None
        ),
    }
    kwargs = {k: v for k, v in kwargs.items() if v is not None}

    if config.pop("post_processing_disable", False):
        kwargs["skip"] = True

    config["post_process_config"] = PostProcessConfig(**kwargs)


def output_options(func):
    """Decorate a click command to add output-specific options."""

    @optgroup.group(
        "Output parameters",
        help="Parameters used to configure the outputs",
    )
    @optgroup.option(
        "--output-dir",
        type=click.Path(file_okay=False),
        required=True,
        help="The directory where the outputs will be stored.",
    )
    @optgroup.option(
        "--outputs-enable-final-figures/--outputs-disable-final-figures",
        default=None,
        help="If enabled, the final figures are exported",
    )
    @optgroup.option(
        "--outputs-enable-graph-creation-figures/--outputs-disable-graph-creation-figures",
        default=None,
        help="If enabled, the graph creation figures are exported",
    )
    @optgroup.option(
        "--outputs-enable-graph-creation-data/--outputs-disable-graph-creation-data",
        default=None,
        help="If enabled, the graph creation data are exported",
    )
    @optgroup.option(
        "--outputs-enable-main-trunk-figures/--outputs-disable-main-trunk-figures",
        default=None,
        help="If enabled, the main trunk figures are exported",
    )
    @optgroup.option(
        "--outputs-enable-main-trunk-morphologies/--outputs-disable-main-trunk-morphologies",
        default=None,
        help="If enabled, the main trunk morphologies are exported",
    )
    @optgroup.option(
        "--outputs-enable-morphologies/--outputs-disable-morphologies",
        default=None,
        help="If enabled, the morphologies with new axons are exported",
    )
    @optgroup.option(
        "--outputs-enable-morphologies-edges/--outputs-disable-morphologies-edges",
        default=None,
        help="If enabled, the morphologies with new axons are exported as edges",
    )
    @optgroup.option(
        "--outputs-enable-postprocess-trunk-figures/--outputs-disable-postprocess-trunk-figures",
        default=None,
        help="If enabled, the post-process trunk figures are exported",
    )
    @optgroup.option(
        "--outputs-enable-postprocess-trunk-morphologies/--outputs-disable-postprocess-trunk-morphologies",
        default=None,
        help="If enabled, the post-process trunk morphologies are exported",
    )
    @optgroup.option(
        "--outputs-enable-steiner-tree-solutions/--outputs-disable-steiner-tree-solutions",
        default=None,
        help="If enabled, the Steiner tree solutions are exported",
    )
    @optgroup.option(
        "--outputs-enable-steiner-tree-solution-figures/--outputs-disable-steiner-tree-solution-figures",
        default=None,
        help="If enabled, the Steiner tree solution figures are exported",
    )
    @optgroup.option(
        "--outputs-enable-target-point-figures/--outputs-disable-target-point-figures",
        default=None,
        help="If enabled, the target point figures are exported",
    )
    @optgroup.option(
        "--outputs-enable-target-points/--outputs-disable-target-points",
        default=None,
        help="If enabled, the target point data are exported",
    )
    @optgroup.option(
        "--outputs-enable-tuft-figures/--outputs-disable-tuft-figures",
        default=None,
        help="If enabled, the tuft figures are exported",
    )
    @optgroup.option(
        "--outputs-enable-tuft-morphologies/--outputs-disable-tuft-morphologies",
        default=None,
        help="If enabled, the tuft morphologies are exported",
    )
    @functools.wraps(func)
    def wrapper_output_options(*args, **kwargs) -> Callable:
        return func(*args, **kwargs)

    return wrapper_output_options


def outputs_kwargs_to_config(config) -> None:
    """Extract the atlas arguments from given config to create an AtlasConfig object."""
    kwargs = {"path": config.pop("output_dir")}

    for k in fields_dict(OutputConfig):  # pylint: disable=not-an-iterable
        name = "outputs_enable_" + k
        if config.pop(name, False):
            kwargs[k] = True

    config["output_config"] = OutputConfig(**kwargs)


def synthesis_options(func):
    """Decorate a click command to add synthesis-specific options."""

    @optgroup.group(
        "Synthesis parameters",
        help="Parameters used to configure the synthesis process",
    )
    @optgroup.option(
        "--input-dir",
        type=click.Path(exists=True, file_okay=False),
        required=True,
        help="The directory containing the inputs.",
    )
    @optgroup.option(
        "--morphology-dir",
        type=click.Path(exists=True, file_okay=False),
        required=True,
        help="The directory containing the input morphologies",
    )
    @optgroup.option(
        "--morphology-data-file",
        type=click.Path(exists=True, dir_okay=False),
        required=True,
        help="The MVD3 or SONATA file containing morphology data.",
    )
    @optgroup.option(
        "-r/-nr",
        "--rebuild-existing-axons/--no-rebuild-existing-axons",
        default=False,
        help="Force rebuilding existing axons.",
    )
    @optgroup.option(
        "--axon-grafting-points-file",
        type=click.Path(exists=True, dir_okay=False),
        required=False,
        help=(
            "Path to the HDF5 file containing the section IDs where the axons should be grafted in "
            "the input morphologies (axons are grafted to the soma if not provided)."
        ),
    )
    @optgroup.option(
        "--population-probabilities-file",
        type=click.Path(exists=True, dir_okay=False),
        required=False,
        help="Path to the file containing the population probabilities.",
    )
    @optgroup.option(
        "--projection-probabilities-file",
        type=click.Path(exists=True, dir_okay=False),
        required=False,
        help="Path to the file containing the projection probabilities.",
    )
    @optgroup.option(
        "--population-tuft-number-file",
        type=click.Path(exists=True, dir_okay=False),
        required=False,
        help="Path to the file containing the tuft number distribution for each target population.",
    )
    @optgroup.option(
        "--trunk-properties-file",
        type=click.Path(exists=True, dir_okay=False),
        required=False,
        help="Path to the file containing the trunk properties.",
    )
    @optgroup.option(
        "--tuft-properties-file",
        type=click.Path(exists=True, dir_okay=False),
        required=False,
        help="Path to the file containing the tuft barcodes given to NeuroTS.",
    )
    @optgroup.option(
        "--tuft-distributions-file",
        type=click.Path(exists=True, dir_okay=False),
        required=False,
        help="Path to the file containing the tuft distributions given to NeuroTS.",
    )
    @optgroup.option(
        "--tuft-parameters-file",
        type=click.Path(exists=True, dir_okay=False),
        required=False,
        help="Path to the file containing the tuft parameters given to NeuroTS.",
    )
    @optgroup.option(
        "--tuft-boundary-max-distance",
        type=click.FloatRange(min=0),
        required=False,
        help=(
            "Maximum distance used for the calculation of the attenuation component near the "
            "boundary."
        ),
    )
    @optgroup.option(
        "--tuft-boundary-scale-coeff",
        type=click.FloatRange(min=0),
        required=False,
        help="Coefficient used in the calculation of the attenuation component near the boundary.",
    )
    @optgroup.option(
        "--target-max-tries",
        type=click.IntRange(min=1),
        required=False,
        help="The maximum number of tries for picking target points.",
    )
    @functools.wraps(func)
    def wrapper_synthesis_options(*args, **kwargs) -> Callable:
        return func(*args, **kwargs)

    return wrapper_synthesis_options


def synthesis_kwargs_to_config(config) -> None:
    """Extract the synthesis arguments from given config to create an SynthesisConfig object."""
    kwargs = {}
    for k in fields_dict(SynthesisConfig):  # pylint: disable=not-an-iterable
        name = k
        value = config.pop(name, None)
        if value is not None:
            kwargs[k] = value

    config["config"] = SynthesisConfig(**kwargs)


@click.command(short_help="Synthesize the axons for the given morphologies")
@synthesis_options
@output_options
@atlas_options(required=True)
@create_graph_options
@post_process_options
@parallel_options
@click.pass_obj
def synthesize(global_config: GlobalConfig, **kwargs):
    """The command to synthesize axons."""
    global_config.to_config(kwargs)
    kwargs.pop("debug", None)
    outputs_kwargs_to_config(kwargs)
    synthesis_kwargs_to_config(kwargs)
    atlas_kwargs_to_config(kwargs)
    create_graph_kwargs_to_config(kwargs)
    post_process_kwargs_to_config(kwargs)
    parallel_kwargs_to_config(kwargs)
    synthesize_axons(**kwargs)



"""Some utils for clustering."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import logging
from collections.abc import Iterator
from collections.abc import Mapping
from copy import deepcopy
from pathlib import Path

import networkx as nx
import numpy as np
import pandas as pd
from jsonschema import Draft7Validator
from jsonschema import ValidationError
from jsonschema import validators
from jsonschema.protocols import Validator
from morphio import IterType
from morphio import PointLevel
from morphio.mut import Morphology as MorphIoMorphology
from neurom import COLS
from neurom import NeuriteType
from neurom.core import Morphology
from neurom.core import Section
from neurom.features.section import section_path_length
from neurom.morphmath import section_length
from tmd.io.conversion import convert_morphio_trees
from tmd.Topology.methods import tree_to_property_barcode
from tmd.Topology.persistent_properties import PersistentAngles
from voxcell import VoxcellError
from voxcell import VoxelData
from voxcell.voxel_data import OrientationField

from axon_synthesis.constants import COORDS_COLS
from axon_synthesis.constants import DEFAULT_POPULATION
from axon_synthesis.typing import FileType
from axon_synthesis.typing import SeedType
from axon_synthesis.utils import disable_loggers
from axon_synthesis.utils import get_axons
from axon_synthesis.utils import keep_only_neurites
from axon_synthesis.utils import save_morphology
from axon_synthesis.utils import sublogger


@disable_loggers("morph_tool.converter")
def export_morph(root_path, morph_name, morph, morph_type, suffix="", logger=None):
    """Export the given morphology to the given path."""
    morph_path = root_path / morph_name
    morph_path = morph_path.with_stem(morph_path.stem + suffix).with_suffix(".asc")
    msg = f"Export {morph_type} morphology to {morph_path}"
    save_morphology(morph, morph_path, msg, logger=logger)
    return morph_path


def compute_shortest_paths(directed_graph, attribute_name=None, source=-1):
    """Compute the shortest paths using an attribute mapping."""
    shortest_paths = nx.single_source_shortest_path(directed_graph, source)

    if attribute_name is None:
        return shortest_paths

    node_mapping = nx.get_node_attributes(directed_graph, attribute_name)
    mapped_shortest_paths = {}
    for k, v in shortest_paths.items():
        mapped_shortest_paths[node_mapping[k]] = [node_mapping[i] for i in v]
    return mapped_shortest_paths


def common_path(graph, nodes, source=None, shortest_paths=None) -> list[int]:
    """Compute the common paths of the given nodes.

    Source should be given only if the graph if undirected.
    Shortest paths can be given if they were already computed before.

    .. warning:: The graph must have only one component.
    """
    if not isinstance(graph, nx.DiGraph) and source is None and shortest_paths is None:
        msg = (
            "Either the source or the pre-computed shortest paths must be provided when using "
            "an undirected graph."
        )
        raise ValueError(msg)

    if shortest_paths is None:
        if isinstance(graph, nx.DiGraph):
            try:
                sources = [k for k, v in graph.in_degree if v == 0]
                if len(sources) > 1:
                    msg = "Several roots found in the directed graph."
                    raise RuntimeError(msg)
                source = sources[0]
            except IndexError as exc:
                # pylint: disable=raise-missing-from
                msg = "Could not find the root of the directed graph."
                raise RuntimeError(msg) from exc
        shortest_paths = nx.single_source_shortest_path(graph, source)

    # Compute the common ancestor
    common_nodes: set[int] = set(shortest_paths[nodes[0]])
    for i in nodes[1:]:
        common_nodes.intersection_update(set(shortest_paths[i]))

    return [i for i in shortest_paths[nodes[0]] if i in common_nodes]


def create_tuft_morphology(
    morph,
    tuft_node_ids,
    common_ancestor_section_id,
    cluster_common_path,
    nodes_to_section_ids,
    shortest_paths,
):
    """Create a new morphology containing only the given tuft."""
    tuft_morph = Morphology(morph)
    for i in tuft_morph.root_sections:
        if i.type != NeuriteType.axon:
            tuft_morph.delete_section(i)

    tuft_nodes = {
        j
        for terminal_id, path in shortest_paths.items()
        if terminal_id in tuft_node_ids
        for j in path
    }.difference(cluster_common_path)

    tuft_sections = {nodes_to_section_ids[i] for i in tuft_nodes}

    for i in tuft_morph.sections:
        if i.id not in tuft_sections:
            tuft_morph.delete_section(i.morphio_section, recursive=False)

    tuft_ancestor = tuft_morph.section(common_ancestor_section_id)

    return tuft_morph, tuft_ancestor


def get_barcode(morph, metric="path_distance", tree_index=0):
    """Compute the barcode of the given morphology."""
    tmd_axon = list(convert_morphio_trees(MorphIoMorphology(morph).as_immutable()))[tree_index]
    tuft_barcode, _ = tree_to_property_barcode(
        tmd_axon,
        lambda tree: tree.get_point_path_distances()
        if metric == "path_distance"
        else tree.get_point_radial_distances(point=morph.soma.center),
        property_class=PersistentAngles,
    )
    return tuft_barcode


def resize_root_section(tuft_morph, tuft_orientation, root_section_idx=0):
    """Resize the root section to 1um."""
    new_root_section = tuft_morph.root_sections[root_section_idx]
    new_root_section.points = np.vstack(
        [
            new_root_section.points[-1] - tuft_orientation,
            new_root_section.points[-1],
        ],
    )
    new_root_section.diameters = np.repeat(new_root_section.diameters[1], 2)


def tuft_morph_path(root_path, group_name, axon_id, tuft_id):
    """Create a tuft file path according to the group name, axon ID and cluster ID."""
    return root_path / f"{Path(group_name).with_suffix('').name}_{axon_id}_{tuft_id}.asc"


def clusters_basic_tuft(
    config_name,
    axon_id,
    group_name,
    group_path,
    group,
    nodes,
    **_kwargs,
):
    """Create one cluster with all points of the group."""
    terminals = nodes.drop(-1)
    new_terminal_points = [
        [
            group_name,
            group_path,
            config_name,
            axon_id,
            1,
            len(terminals),
            *terminals[COORDS_COLS].mean().tolist(),
        ]
    ]

    group.loc[
        group["graph_node_id"].isin(terminals.index),
        "tuft_id",
    ] = 1

    return new_terminal_points, group["tuft_id"]


def compute_mean_tuft_length(
    volume,
    n_pot,
    fraction,
    strength,
    bouton_density,
):
    """Compute properties for the cluster."""
    n_act = n_pot * fraction
    n_tot = volume * strength
    n_syn_per = n_tot / n_act
    return n_syn_per / bouton_density


def compute_common_section_properties(
    first_axon_pt, section, n_pot, volume, fraction, strength, bouton_density
):
    """Compute basic properties of the given section."""
    path_distance = sum(section_length(i.points) for i in section.iter(IterType.upstream))
    radial_distance = np.linalg.norm(first_axon_pt - section.points[-1])
    path_length = sum(section_length(i.points) for i in section.iter())
    max_path_extent = max(map(section_path_length, Section.ileaf(Section(section))))

    if all(i is not None for i in [n_pot, volume, fraction, strength, bouton_density]):
        mean_tuft_length = compute_mean_tuft_length(
            volume,
            n_pot,
            fraction,
            strength,
            bouton_density,
        )
    else:
        mean_tuft_length = None

    return path_distance, radial_distance, path_length, mean_tuft_length, max_path_extent


def reduce_clusters(  # noqa: C901, PLR0912, PLR0913, PLR0915
    group,
    group_path,
    group_name,
    morph,
    axon,
    axon_id,
    cluster_df,
    directed_graph,
    nodes,
    sections_to_add,
    morph_paths,
    cluster_props: list[tuple],
    shortest_paths: dict,
    bouton_density: float | None,
    brain_regions: VoxelData | None = None,
    atlas_region_id: int | None = None,
    atlas_orientations: OrientationField | None = None,
    projection_pop_numbers: pd.DataFrame | None = None,
    export_tuft_morph_dir: FileType | None = None,
    config_name: str | None = None,
    rng: SeedType = None,
    logger: logging.Logger | logging.LoggerAdapter | None = None,
) -> set[int]:
    """Reduce clusters to one section from their common ancestors to their centers."""
    # pylint: disable=too-many-branches, too-many-statements
    logger = sublogger(logger, __name__)
    if not config_name:
        config_name = ""
    kept_path: set[int] = set()
    group = group.dropna(subset="tuft_id").astype({"tuft_id": int})

    root_point = axon.points[0, COLS.XYZ]

    if projection_pop_numbers is not None:
        source_projections = projection_pop_numbers.loc[
            projection_pop_numbers["atlas_region_id"] == atlas_region_id
        ]
    else:
        source_projections = pd.DataFrame(columns=["target_atlas_id"])

    for tuft_id, cluster in group.groupby("tuft_id"):
        # Skip the root cluster
        if (cluster.tuft_id <= 0).any():
            continue

        # Compute the common ancestor of the nodes
        cluster_common_path = common_path(
            directed_graph,
            cluster["graph_node_id"].tolist(),
            shortest_paths=shortest_paths,
        )
        common_ancestor_shift = -2 if len(cluster) == 1 and len(cluster_common_path) > 2 else -1
        common_ancestor = nodes.loc[cluster_common_path[common_ancestor_shift], "section_id"]
        common_section = morph.section(common_ancestor)

        kept_path = kept_path.union(nodes.loc[cluster_common_path, "section_id"])

        # Create a morphology for the current tuft
        tuft_morph, tuft_ancestor = create_tuft_morphology(
            morph,
            set(cluster["graph_node_id"]),
            common_ancestor,
            cluster_common_path[:common_ancestor_shift],
            nodes["section_id"].to_dict(),
            shortest_paths,
        )

        # Compute cluster center
        cluster_center = cluster_df.loc[
            cluster_df["terminal_id"] == tuft_id,
            COORDS_COLS,
        ].to_numpy()[0]

        # Compute tuft orientation
        ancestor_pts = (
            tuft_ancestor.points[-1]
            if not common_section.is_root or common_section.children
            else tuft_ancestor.points.mean(axis=0)
        )
        tuft_orientation = cluster_center - ancestor_pts
        tuft_orientation /= np.linalg.norm(tuft_orientation)
        if atlas_orientations is not None:
            try:
                atlas_orientation = atlas_orientations.lookup(ancestor_pts)[0].T
            except VoxcellError:
                logger.debug(
                    "Could not retrieve the atlas orientation for %s at %s",
                    group_name,
                    ancestor_pts,
                )
            else:
                if ~np.isnan(atlas_orientation).any():
                    tuft_orientation = np.dot(tuft_orientation, atlas_orientation)

        # Resize the common section used as root (the root section is 1um)
        resize_root_section(tuft_morph, tuft_orientation)

        # Compute the barcode
        tuft_barcode = get_barcode(tuft_morph)

        if export_tuft_morph_dir is not None:
            # Export each tuft as a morphology
            morph_paths["tufts"].append(
                (
                    group_name,
                    group_path,
                    config_name,
                    axon_id,
                    tuft_id,
                    export_morph(
                        export_tuft_morph_dir,
                        group_name,
                        tuft_morph,
                        "tuft",
                        f"_{config_name}_{axon_id}_{tuft_id}",
                        logger=logger,
                    ),
                ),
            )

        # Add tuft category data
        # TODO: Fix the WMR stuff
        try:
            target_atlas_region_id = brain_regions.lookup(  # type: ignore[union-attr]
                common_section.points[-1]
            )
        except:  # noqa: E722 ; pylint: disable=bare-except
            target_atlas_region_id = 0
        target_projection_number = source_projections.loc[
            source_projections["target_atlas_id"] == target_atlas_region_id
        ]
        if len(target_projection_number) > 0:
            target_tmp = target_projection_number.sample(random_state=rng)
            target_projection_number, volume, target_projection_name = target_tmp[
                ["pop_neuron_numbers", "atlas_region_volume_target", "target_projection_name"]
            ].to_numpy()[0]
            fraction = None  # TODO: Get proper values
            strength = None  # TODO: Get proper values
        else:
            target_projection_number = None
            target_projection_name = DEFAULT_POPULATION
            fraction = None
            strength = None
            volume = None

        (
            path_distance,
            radial_distance,
            path_length,
            mean_tuft_length,
            max_path_extent,
        ) = compute_common_section_properties(
            root_point,
            common_section,
            target_projection_number,
            volume,
            fraction,
            strength,
            bouton_density,
        )

        # TODO: Compute the cluster weights somewhere else

        cluster_props.append(
            (
                group_name,
                group_path,
                config_name,
                axon_id,
                tuft_id,
                cluster_center.tolist(),
                common_ancestor,
                *ancestor_pts.tolist(),
                path_distance,
                radial_distance,
                path_length,
                max_path_extent,
                len(cluster),
                tuft_orientation.tolist(),
                mean_tuft_length,
                target_atlas_region_id,
                target_projection_name,
                np.array(tuft_barcode).tolist(),
            ),
        )

        # Continue if the cluster has only one node
        if len(cluster) == 1:
            continue

        # Create a new section from the common ancestor to the center of the cluster
        sections_to_add[common_section.id].append(
            PointLevel(
                [
                    common_section.points[-1],
                    cluster_df.loc[
                        (cluster_df["terminal_id"] == tuft_id),
                        COORDS_COLS,
                    ].to_numpy()[0],
                ],
                [0, 0],
            ),
        )

    if (group["tuft_id"] == -1).any():
        kept_path = kept_path.union(group.loc[group["tuft_id"] == -1, "section_id"])
    return kept_path


def create_clustered_morphology(
    morph, group_name, axon_id, kept_path, sections_to_add, suffix=None
):
    """Create a new morphology with the kept path and add new sections to cluster centers."""
    if not suffix:
        suffix = ""
    clustered_morph = Morphology(
        morph,
        name=f"Clustered {group_name}{suffix}",
    )
    trunk_morph = Morphology(
        clustered_morph,
        name=f"Clustered trunk {group_name}{suffix}",
    )

    axon = get_axons(morph, axon_id)
    new_axon = get_axons(clustered_morph, axon_id)
    trunk_axon = get_axons(trunk_morph, axon_id)

    root = axon.root_node
    new_root = new_axon.root_node
    new_trunk_root = trunk_axon.root_node

    if not np.array_equal(root.points, new_root.points):
        msg = "The axons were messed up!"
        raise RuntimeError(msg)

    for sec in new_root.children:
        clustered_morph.delete_section(sec.morphio_section)
    for sec in new_trunk_root.children:
        trunk_morph.delete_section(sec.morphio_section)

    current_sections = [(root, new_root, new_trunk_root)]

    # Add kept sections
    while current_sections:
        current_section, current_new_section, current_trunk_section = current_sections.pop()
        for child in current_section.children:
            if child.id in kept_path:
                new_section = PointLevel(
                    child.points[:, COLS.XYZ].tolist(),
                    (child.points[:, COLS.R] * 2).tolist(),
                )
                current_sections.append(
                    (
                        child,
                        current_new_section.append_section(new_section),
                        current_trunk_section.append_section(new_section),
                    ),
                )

        if current_section.id in sections_to_add:
            for new_sec in sections_to_add[current_section.id]:
                current_new_section.append_section(new_sec)

    for i in [clustered_morph, trunk_morph]:
        keep_only_neurites(i, NeuriteType.axon, axon_id)
    return clustered_morph, trunk_morph


def extend_validator_with_default(validator_class) -> Validator:
    """Extend a validator to automatically set default values during validation."""
    _no_default = object()
    validate_properties = validator_class.VALIDATORS["properties"]

    def set_defaults_and_validate(  # type: ignore[return]
        validator, properties, instance, schema
    ) -> Iterator[ValidationError] | None:
        drop_if_empty = set()
        new_instance = deepcopy(instance)
        for prop, subschema in properties.items():
            if prop in new_instance:
                continue
            obj_type = subschema.get("type", "")
            default_value = subschema.get("default", _no_default)
            if default_value is not _no_default:
                new_instance.setdefault(prop, default_value)
            elif obj_type == "object":
                new_instance.setdefault(prop, {})
                drop_if_empty.add(prop)

        is_valid = True
        for error in validate_properties(
            validator,
            properties,
            new_instance,
            schema,
        ):
            is_valid = False
            yield error

        for prop in drop_if_empty:
            instance_prop = new_instance[prop]
            if isinstance(instance_prop, Mapping) and len(instance_prop) == 0:
                del new_instance[prop]

        if is_valid:
            instance.update(new_instance)

    return validators.extend(
        validator_class,
        {"properties": set_defaults_and_validate},
    )


DefaultValidatingValidator = extend_validator_with_default(Draft7Validator)



"""Common entries of the Command Line Interface."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
# LICENSE HEADER MANAGED BY add-license-header

import functools
from collections.abc import Callable

import click
from click_option_group import optgroup

from axon_synthesis.atlas import AtlasConfig
from axon_synthesis.cli.utils import DictParam
from axon_synthesis.cli.utils import ListParam
from axon_synthesis.utils import ParallelConfig


def atlas_options(*, required=False):
    """Decorate a click command to add Atlas-specific options."""

    def inner_atlas_options(func) -> Callable:
        @optgroup.group("Atlas parameters", help="Parameters used to read and prepare the atlas")
        @optgroup.option(
            "--atlas-enable/--atlas-disable",
            default=True,
            help="Enable or disable the Atlas",
        )
        @optgroup.option(
            "--atlas-path",
            type=click.Path(exists=True, file_okay=False),
            required=required,
            help="Path to the Atlas folder",
        )
        @optgroup.option(
            "--atlas-region-filename",
            type=str,
            help="Name of NRRD file containing the brain regions in the Atlas folder",
        )
        @optgroup.option(
            "--atlas-layer-names",
            type=ListParam(),
            help=(
                "Names of the layers given as a JSON list (the atlas folder must contain a file "
                "name '[PH]<layer_name>.nrrd' for each given layer)"
            ),
        )
        @optgroup.option(
            "--atlas-outside-region-id",
            type=int,
            help="ID of the 'outside' brain region",
        )
        @optgroup.option(
            "--atlas-enable-boundary/--atlas-disable-boundary",
            default=None,
            help="Enable or disable the boundary conditions",
        )
        @functools.wraps(func)
        def wrapper_atlas_options(*args, **kwargs) -> Callable:
            return func(*args, **kwargs)

        return wrapper_atlas_options

    return inner_atlas_options


def atlas_kwargs_to_config(config) -> None:
    """Extract the atlas arguments from given config to create an AtlasConfig object."""
    kwargs = {
        k: v
        for k, v in {
            "path": config.pop("atlas_path", None),
            "region_filename": config.pop("atlas_region_filename", None),
            "layer_names": config.pop("atlas_layer_names", None),
            "outside_region_id": config.pop("atlas_outside_region_id", None),
            "use_boundary": config.pop("atlas_enable_boundary", None),
        }.items()
        if v is not None
    }
    if config.pop("atlas_enable", False):
        config["atlas_config"] = AtlasConfig(**kwargs)
    else:
        config["atlas_config"] = None


def parallel_options(func):
    """Decorate a click command to add parallel-specific options."""

    @optgroup.group(
        "Parallel computation parameters",
        help="Parameters used to configure the parallel computation",
    )
    @optgroup.option(
        "--nb-workers",
        type=click.IntRange(min=0),
        help="The number of workers",
    )
    @optgroup.option(
        "--dask-config",
        type=DictParam(),
        help="The dask configuration given as a JSON string.",
    )
    @optgroup.option(
        "--progress-bar/--no-progress-bar",
        default=None,
        help="If set to True, a progress bar is displayed during computation",
    )
    @optgroup.option(
        "--use-mpi/--no-use-mpi",
        default=None,
        help="If set to True, MPI is used for parallel computation",
    )
    @functools.wraps(func)
    def wrapper_parallel_options(*args, **kwargs) -> Callable:
        return func(*args, **kwargs)

    return wrapper_parallel_options


def parallel_kwargs_to_config(config) -> None:
    """Extract the parallel arguments from given config to create an ParallelConfig object."""
    kwargs = {
        "nb_processes": config.pop("nb_workers", None),
        "dask_config": config.pop("dask_config", None),
        "progress_bar": config.pop("progress_bar", None),
        "use_mpi": config.pop("use_mpi", None),
    }
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    config["parallel_config"] = ParallelConfig(**kwargs)



"""Entries of the Command Line Interface dedicated to the scripts."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
# LICENSE HEADER MANAGED BY add-license-header

import logging

import click

from axon_synthesis.atlas import AtlasHelper
from axon_synthesis.cli.synthesis import atlas_kwargs_to_config
from axon_synthesis.cli.synthesis import atlas_options
from axon_synthesis.cli.utils import GlobalConfig
from axon_synthesis.cli.utils import ListParam
from axon_synthesis.utils import create_random_morphologies


@click.command(
    short_help="Create empty cells at random locations in the atlas.",
)
@click.option(
    "--nb-morphologies",
    type=click.IntRange(min=1),
    required=True,
    help="The number of empty morphologies to generate",
)
@click.option(
    "--brain-regions",
    type=ListParam(),
    help="The list of brain regions in which the empty morphologies should be located",
)
@click.option(
    "--output-morphology-dir",
    type=click.Path(file_okay=False),
    help="The directory where the morphologies will be exported",
)
@click.option(
    "--output-cell-collection",
    type=click.Path(dir_okay=False),
    help="The path to the output cell collection",
)
@click.option(
    "--morphology-prefix",
    type=str,
    help="The prefix of morphology names",
)
@atlas_options(required=True)
@click.pass_obj
def random_morphologies(global_config: GlobalConfig, *_args, **kwargs):
    """The command to create random empty morphologies."""
    global_config.to_config(kwargs)
    kwargs.pop("debug", None)
    atlas_kwargs_to_config(kwargs)
    kwargs["atlas_config"].load_region_map = True
    kwargs["atlas"] = AtlasHelper(kwargs.pop("atlas_config"))
    kwargs["logger"] = logging.getLogger(__name__)
    create_random_morphologies(**kwargs)



"""Configuration for the pytest test suite."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

# pylint: disable=redefined-outer-name
import logging
import shutil
from pathlib import Path

import dir_content_diff.comparators.pandas
import pytest
from voxcell.nexus.voxelbrain import Atlas

from . import DATA
from . import EXAMPLES
from . import TEST_ROOT
from .data_factories import generate_small_O1

logging.getLogger("matplotlib").disabled = True
logging.getLogger("matplotlib.font_manager").disabled = True

dir_content_diff.comparators.pandas.register()


def pytest_addoption(parser):
    """Hook to add custom options to the CLI of pytest."""
    parser.addoption(
        "--interactive-plots",
        action="store_true",
        default=False,
        help="Trigger interactive plots in tests to check the results",
    )


@pytest.fixture
def interactive_plots(request):
    """The value given to the option for interactive plots."""
    return request.config.getoption("--interactive-plots")


def pytest_configure(config):
    """Add --check-untyped-defs option to the mypy plugin."""
    plugin = config.pluginmanager.getplugin("mypy")
    plugin.mypy_argv.append("--check-untyped-defs")


@pytest.fixture
def root_dir():
    """The root directory."""
    return Path(TEST_ROOT)


@pytest.fixture
def data_dir():
    """The data directory."""
    return Path(DATA)


@pytest.fixture
def example_dir():
    """The example directory."""
    return Path(EXAMPLES)


@pytest.fixture
def testing_dir(tmpdir, monkeypatch):
    """The testing directory."""
    monkeypatch.chdir(tmpdir)
    return Path(tmpdir)


@pytest.fixture
def out_dir(testing_dir):
    """The output directory."""
    path = testing_dir / "out"
    path.mkdir(parents=True)

    return path


@pytest.fixture(scope="session")
def atlas_path(tmpdir_factory):
    """Generate a small O1 atlas for the test session."""
    atlas_directory = tmpdir_factory.mktemp("atlas_small_O1")
    return generate_small_O1(atlas_directory)


@pytest.fixture
def atlas(atlas_path):
    """Load the small O1 atlas."""
    return Atlas.open(str(atlas_path))


@pytest.fixture
def brain_regions(atlas):
    """Load the brain regions of the small O1 atlas."""
    return atlas.load_data("brain_regions")


@pytest.fixture(scope="session")
def morphology_path(tmpdir_factory):
    """Generate a small O1 atlas for the test session."""
    morph_directory = Path(tmpdir_factory.mktemp("morphologies"))
    for i in (DATA / "input_morphologies").iterdir():
        shutil.copyfile(i, morph_directory / i.name)
    return morph_directory


@pytest.fixture
def wmr_path(out_dir):
    """Generate a white matter recipe file.

    This WMR is compatible with the small O1 atlas.
    """
    wmr_filepath = out_dir / "white_matter_recipe" / "white_matter_recipe_test.yaml"
    wmr_filepath.parent.mkdir(parents=True, exist_ok=True)
    shutil.copyfile(DATA / "white_matter_recipe.yaml", wmr_filepath)
    return wmr_filepath


@pytest.fixture
def _tuft_inputs(testing_dir) -> None:
    """Copy inputs for tuft generation in the testing directory."""
    shutil.copyfile(DATA / "tuft_distributions.json", testing_dir / "tuft_distributions.json")
    shutil.copyfile(DATA / "tuft_parameters.json", testing_dir / "tuft_parameters.json")



"""Some utils for the CLI of axon-synthesis."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
# LICENSE HEADER MANAGED BY add-license-header

import json
from collections.abc import MutableMapping
from copy import deepcopy
from typing import Any

import click
from configobj import ConfigObj
from jsonschema import validate


def _format_value(data: dict, name: str) -> dict[str, Any]:
    return {f"{name}_{key}": value for key, value in data.items()}


def _recursive_merge(dict_1: MutableMapping, dict_2: MutableMapping) -> MutableMapping:
    """Merge two dictionaries recursively.

    The right one takes precedense in case of conflicting keys.
    """
    merged_dict = deepcopy(dict_1)
    for k, v in dict_2.items():
        if k in merged_dict and isinstance(v, MutableMapping):
            merged_dict[k] = _recursive_merge(merged_dict[k], v)
        else:
            merged_dict[k] = v
    return merged_dict


def _flatten_command_subsections(ctx, command_group, command_defaults) -> None:
    for command_name in command_group.list_commands(ctx):
        command = command_group.get_command(ctx, command_name)
        if isinstance(command, click.core.Group):
            _flatten_command_subsections(ctx, command, command_defaults[command_name])
        else:
            to_add = {}
            to_remove = []
            for subkey, subvalue in command_defaults[command_name].items():
                if isinstance(subvalue, dict):
                    to_add.update(_format_value(subvalue, subkey))
                    to_remove.append(subkey)
            command_defaults[command_name] = _recursive_merge(
                to_add, command_defaults[command_name]
            )
            for i in to_remove:
                del command_defaults[command_name][i]


def _process_command(ctx, command, defaults, global_values) -> None:
    for subcommand_name in command.list_commands(ctx):
        subcommand = command.get_command(ctx, subcommand_name)
        if subcommand_name not in defaults:
            defaults[subcommand_name] = {}
        if isinstance(subcommand, click.core.Group):
            _process_command(ctx, subcommand, defaults[subcommand_name], global_values)
            continue
        defaults[subcommand_name] = _recursive_merge(global_values, defaults[subcommand_name])


def configure(ctx: click.Context, _, filename: None | str):
    """Set parameter default values according to a given configuration file."""
    if filename is None:
        return

    # Load the config file
    cfg = ConfigObj(filename)

    # Get current default values
    defaults = cfg.dict()

    # Copy global arguments to all sub commands
    global_values = defaults.get("global", None)

    if global_values:
        _process_command(ctx, ctx.command, defaults, global_values)

    # Flatten sub-sections
    _flatten_command_subsections(ctx, ctx.command, defaults)

    ctx.default_map = defaults


class GlobalConfig:
    """Class to store global configuration."""

    def __init__(self, *, debug=False, seed=None):
        """The GlobalConfig constructor."""
        self.debug = debug
        self.seed = seed

    def to_config(self, config):
        """Copy internal attributes in the given dictionary."""
        config["debug"] = self.debug
        config["rng"] = self.seed


class ListParam(click.ParamType):
    """A `click` parameter to process parameters given as JSON arrays."""

    name = "list"

    def __init__(self, *args, schema=None, **kwargs):
        """The ListParam constructor."""
        self.schema = schema
        super().__init__(*args, **kwargs)

    def convert(self, value, param, ctx):
        """Convert a given value."""
        try:
            if not isinstance(value, list):
                value = json.loads(value)
        except json.JSONDecodeError:
            self.fail(f"{value!r} is not a valid JSON array", param, ctx)
        if self.schema is not None:
            validate(value, schema=self.schema)
        return value


class DictParam(click.ParamType):
    """A `click` parameter to process parameters given as JSON objects."""

    name = "dict"

    def __init__(self, *args, schema=None, **kwargs):
        """The DictParam constructor."""
        self.schema = schema
        super().__init__(*args, **kwargs)

    def convert(self, value, param, ctx):
        """Convert a given value."""
        try:
            if not isinstance(value, dict):
                value = json.loads(value)
        except json.JSONDecodeError:
            self.fail(f"{value!r} is not a valid JSON object", param, ctx)
        if self.schema is not None:
            validate(value, schema=self.schema)

        return value



"""Tests for the statistics functions."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

from diff_pdf_visually import pdf_similar
from morph_tool.utils import iter_morphology_files

from axon_synthesis.validation.statistics import default_config
from axon_synthesis.validation.statistics import plot_score_matrix


def test_plot_score_matrix(data_dir, tmp_path):
    """Test the plot_score_matrix function."""

    def get_morphs(input_dir) -> list:
        return [(i.name, [i]) for i in sorted(iter_morphology_files(input_dir))]

    ref_morphs_files = get_morphs(data_dir / "mimic_example" / "inputs" / "converted_morphologies")
    res_morphs_files = get_morphs(
        data_dir / "mimic_example" / "synthesis_basic_172992" / "Morphologies"
    ) + get_morphs(data_dir / "mimic_example" / "synthesis_basic_172993" / "Morphologies")

    output = tmp_path / "score_matrix.pdf"
    plot_score_matrix(ref_morphs_files, res_morphs_files, output, default_config())

    expected = data_dir / "score_matrix.pdf"
    pdf_similar(output, expected)



"""Compute and plot some statistics."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import logging
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import cm
from matplotlib.backends.backend_pdf import PdfPages
from morph_tool.utils import iter_morphology_files
from neurom import load_morphologies
from neurom.apps.morph_stats import extract_dataframe

# import json
# from collections import defaultdict
# from itertools import chain

# import attr
# import luigi
# import luigi_tools
# import neurom as nm
# import neurots.extract_input
# from data_validation_framework.target import TaggedOutputLocalTarget
# from luigi.parameter import OptionalPathParameter
# from luigi.parameter import PathParameter
# from neurom import load_morphology
# from neurom.core.types import NeuriteType

# from axon_synthesis.add_tufts import AddTufts
# from axon_synthesis.create_dataset import RepairDataset
# from axon_synthesis.main_trunk.clustering import ClusterTerminals

# # from maint_trunk.clustering import ClusterTerminals
# # from maint_trunk.steiner_morphologies import SteinerMorphologies

logger = logging.getLogger(__name__)


# def _np_cast(array, do_sum=False):
#     if do_sum:
#         return np.array(array).sum()
#     return np.array(array).tolist()


def default_config():
    """Create the default config used to compute the scores."""
    return {
        "neurite": {
            "number_of_bifurcations": ["sum"],
            "number_of_sections_per_neurite": ["mean", "sum"],
            "number_of_leaves": ["sum"],
            "partition_asymmetry": ["mean", "sum"],
            "partition_asymmetry_length": ["mean", "sum"],
            "remote_bifurcation_angles": ["mean"],
            "section_bif_branch_orders": ["max", "mean"],
            "section_bif_lengths": ["min", "max", "mean"],
            "section_bif_radial_distances": ["max", "mean"],
            "section_branch_orders": ["max", "mean"],
            "section_lengths": ["min", "max", "mean", "sum"],
            "section_path_distances": ["min", "max", "mean"],
            "section_radial_distances": ["min", "max", "mean"],
            "section_strahler_orders": ["mean"],
            "section_term_branch_orders": ["max", "mean"],
            "section_term_lengths": ["min", "max", "mean"],
            "section_term_radial_distances": ["max", "mean"],
            "section_tortuosity": ["min", "max", "mean"],
            "total_length_per_neurite": ["mean"],
        },
        "neurite_type": [
            "AXON",
        ],
    }


def relative_score(data1, data2):
    """Get the score.

    Args:
        data1 (list): the first data set.
        data2 (list): the second data set.
    """
    return (data2 - data1) / data1


def get_scores(df1, df2):
    """Return scores between two data sets.

    Args:
        df1 (pandas.DataFrame): the first data set.
        df2 (pandas.DataFrame): the second data set.

    Returns:
        The list of feature scores.
    """
    scores = []
    score_names = []
    key_names = {
        "basal_dendrite": "Basal",
        "apical_dendrite": "Apical",
        "axon": "Axon",
    }
    for neurite_type, neurite_name in key_names.items():
        if neurite_type in df1.columns and neurite_type in df2.columns:
            _df1 = df1[neurite_type]
            _df2 = df2[neurite_type]
            for k in _df1.columns:
                data1 = _df1[k].to_numpy()[0]
                data2 = _df2[k].to_numpy()[0]
                score_name = neurite_name + " " + k.replace("_", " ")
                score_names.append(score_name)
                if not np.isnan(data1) and not np.isnan(data2):
                    sc1 = relative_score(data1, data2)
                    if not np.isnan(sc1):
                        scores.append(sc1)
                    else:
                        scores.append(0.0)
                else:
                    scores.append(np.nan)
                logger.debug(
                    "Score name: %s ; Biological value: %s ; Generated value: %s ; Score: %s",
                    score_name,
                    data1,
                    data2,
                    scores[-1],
                )

    return score_names, scores


def compute_scores(ref, test, config):
    """Compute scores of a test population against a reference population.

    Args:
        ref (tuple(str, list)): the reference data.
        test (tuple(str, list)): the test data.
        config (dict): the configuration used to compute the scores.

    Returns:
        The scores and the feature list.
    """
    ref_mtype, ref_files = ref
    test_mtype, test_files = test
    if ref_mtype != test_mtype:
        msg = "The mtypes of ref and test files must be the same."
        raise AssertionError(msg)

    ref_pop = load_morphologies(ref_files)
    test_pop = load_morphologies(test_files)

    ref_features = extract_dataframe(ref_pop, config)
    test_features = extract_dataframe(test_pop, config)

    logger.debug(ref_files)
    return get_scores(ref_features, test_features)


def _check_scores(scores, keys) -> int:
    n_scores = len(keys[0])
    for k, s in zip(keys[1:], scores):
        if keys[0] != k:
            msg = "Score names must all be the same for each feature."
            raise AssertionError(msg)
        if len(k) != n_scores:
            msg = "The number of keys must be the same for each morphology."
            raise AssertionError(msg)
        if len(s) != n_scores:
            msg = "The number of scores must be the same for each morphology."
            raise AssertionError(msg)
    return n_scores


def plot_score_matrix(
    ref_morphs_dir,
    test_morphs_dir,
    output_path,
    config,
    dpi=100,
):
    """Plot score matrix for a test population against a reference population."""
    # pylint: disable=too-many-locals
    Path(output_path).parent.mkdir(parents=True, exist_ok=True)

    # Build the file list
    if not isinstance(ref_morphs_dir, list):
        ref_file_lists = [(i.name, [i]) for i in sorted(iter_morphology_files(ref_morphs_dir))]
    else:
        ref_file_lists = ref_morphs_dir
    if not isinstance(test_morphs_dir, list):
        test_file_lists = [(i.name, [i]) for i in sorted(iter_morphology_files(test_morphs_dir))]
    else:
        test_file_lists = test_morphs_dir
    size = len(ref_file_lists)
    names = [i[0] for i in ref_file_lists]
    if size != len(test_file_lists):
        msg = (
            f"The number of ref files ({size}) must be equal to the res files "
            f"({len(test_file_lists)})"
        )
        raise AssertionError(msg)

    # Compute scores
    scores = []
    keys = []
    for ref_files, test_files in zip(ref_file_lists, test_file_lists):
        key_name, score = compute_scores(ref_files, test_files, config)
        keys.append(key_name)
        scores.append(score)

    n_scores = _check_scores(scores, keys)

    # Plot statistics
    with PdfPages(output_path) as pdf:
        # Compute subplot ratios and figure size
        height_ratios = [7, (1 + n_scores)]
        fig_width = size
        fig_height = sum(height_ratios) * 0.3

        hspace = 0.625 / fig_height
        wspace = 0.2 / fig_width

        cbar_ratio = 0.4 / fig_width

        # Create the figure and the subplots
        fig, ((a0, a2), (a1, a3)) = plt.subplots(  # type: ignore[misc]
            2,
            2,
            gridspec_kw={
                "height_ratios": height_ratios,
                "width_ratios": [1 - cbar_ratio, cbar_ratio],
                "hspace": hspace,
                "wspace": wspace,
            },
            figsize=(fig_width, fig_height),
        )

        # Plot score errors
        clipped_scores = np.clip(np.abs(scores), 0, 1)
        a0.errorbar(
            np.arange(size),
            np.nanmean(clipped_scores, axis=1),
            yerr=np.nanstd(clipped_scores, axis=1),
            color="black",
            label="Synthesized",
        )
        a0.tick_params(bottom=False, top=True, labelbottom=False, labeltop=True)
        a0.xaxis.set_tick_params(rotation=45)
        a0.set_xticks(np.arange(size))
        a0.set_xticklabels(names)

        a0.set_xlim([a0.xaxis.get_ticklocs().min() - 0.5, a0.xaxis.get_ticklocs().max() + 0.5])
        a0.set_ylim([-0.1, 1.1])

        # Plot score heatmap
        scores_transpose = np.transpose(scores)
        scores_df = pd.DataFrame(scores_transpose, index=keys[0], columns=names)

        g = sns.heatmap(
            scores_df,
            vmin=-1,
            vmax=1,
            mask=np.isnan(scores_transpose),
            ax=a1,
            cmap=cm.seismic,  # type: ignore[attr-defined]
            cbar_ax=a3,
        )

        g.xaxis.set_tick_params(rotation=45)
        g.set_facecolor("xkcd:black")

        # Remove upper right subplot
        a2.remove()

        # Export the figure
        try:
            logging.disable(logging.CRITICAL)
            pdf.savefig(fig, bbox_inches="tight", dpi=dpi)
        finally:
            logging.disable(0)
        plt.close(fig)


# @attr.s(auto_attribs=True)
# class Statistics:
#     """The object to store basic statistics."""

#     min: np.number
#     max: np.number
#     mean: np.number
#     std: np.number

#     def to_list(self):
#         """Return a list with the min, max, mean and std stat values."""
#         return [
#             self.min,
#             self.max,
#             self.mean,
#             self.std,
#         ]

#     @staticmethod
#     def gather_stats(stats, prefix=""):
#         """Gather stats from a list of Statistics instances."""
#         values = np.array([i.to_list() for i in stats])
#         return {
#             f"{prefix}min": values[:, 0].tolist(),
#             f"{prefix}max": values[:, 1].tolist(),
#             f"{prefix}mean": values[:, 2].tolist(),
#             f"{prefix}std": values[:, 3].tolist(),
#         }


# def to_stats(values):
#     """Compute statistics of the given list of values."""
#     if isinstance(values[0], list):
#         values = np.array(list(chain.from_iterable(values)))
#     else:
#         values = np.array(values)
#     return Statistics(
#         values.min(),
#         values.max(),
#         values.mean(),
#         values.std(),
#     )


# def population_statistics(pop, neurite_type=NeuriteType.axon):
#     """Compute statistics for the given population."""
#     # Statistics we want to check
#     # section_tortuosity = []
#     # section_radial_distances = []
#     terminal_path_lengths = []
#     # section_term_radial_distances = []
#     # neurite_tortuosity = []
#     local_bifurcation_angles = []
#     remote_bifurcation_angles = []
#     total_axon_length = []
#     radial_moment_0 = []
#     # radial_moment_1 = []
#     radial_moment_2 = []
#     # normalized_radial_moment_0 = []
#     # normalized_radial_moment_1 = []
#     normalized_radial_moment_2 = []
#     for neuron in pop:
#         logger.info(neuron)
#         # import pdb
#         # pdb.set_trace()
#         # neurite_tortuosity.append(
#         #     _np_cast(nm.get("tortuosity_per_neurite", neuron, neurite_type=neurite_type))
#         # )
#         # neurite_tortuosity.append(
#         #     _np_cast(nm.get("tortuosity", neuron, neurite_type=neurite_type))
#         # )
#         # section_tortuosity.append(
#         #     _np_cast(nm.get("section_tortuosity", neuron, neurite_type=neurite_type))
#         # )
#         # section_radial_distances.append(
#         #     _np_cast(nm.get("section_radial_distances", neuron, neurite_type=neurite_type))
#         # )
#         # section_term_radial_distances.append(
#         #     _np_cast(nm.get("section_term_radial_distances", neuron, neurite_type=neurite_type))
#         # )
#         # terminal_path_lengths.append(
#         #     to_stats(nm.get("terminal_path_lengths", neuron, neurite_type=neurite_type))
#         # )
#         # neurite_tortuosity.append((
#         #     np.array(terminal_path_lengths[-1])
#         #     / np.array(section_term_radial_distances[-1])
#         # ).tolist())
#         local_bifurcation_angles.append(
#             to_stats(nm.get("local_bifurcation_angles", neuron, neurite_type=neurite_type)),
#         )
#         remote_bifurcation_angles.append(
#             to_stats(nm.get("remote_bifurcation_angles", neuron, neurite_type=neurite_type)),
#         )
#         total_axon_length.append(
#             sum(nm.get("total_length_per_neurite", neuron, neurite_type=neurite_type)),
#         )
#         radial_moments = {
#             i: nm.get(
#                 "radial_moment",
#                 neuron,
#                 neurite_type=neurite_type,
#                 order=i,
#                 use_radius=False,
#             )
#             for i in [0, 2]
#         }
#         normalized_moments = {order: m / radial_moments[0] for order, m in radial_moments.items()}
#         radial_moment_0.append(radial_moments[0])
#         # radial_moment_1.append(radial_moments[1])
#         radial_moment_2.append(radial_moments[2])
#         # normalized_radial_moment_1.append(normalized_moments[1])
#         normalized_radial_moment_2.append(normalized_moments[2])

#     result = {
#         # "section_tortuosity": section_tortuosity,
#         # "neurite_tortuosity": neurite_tortuosity,
#         # "local_bifurcation_angles_stats": to_stats(local_bifurcation_angles),
#         # "remote_bifurcation_angles_stats": to_stats(remote_bifurcation_angles),
#         # "section_radial_distances": section_radial_distances,
#         # "section_term_radial_distances": to_stats(section_term_radial_distances),
#         # "terminal_path_lengths": to_stats(terminal_path_lengths),
#         "total_axon_length": _np_cast(total_axon_length),
#         "radial_moment_0": radial_moment_0,
#         # "radial_moment_1": radial_moment_1,
#         "radial_moment_2": radial_moment_2,
#         # "normalized_radial_moment_1": normalized_radial_moment_1,
#         "normalized_radial_moment_2": normalized_radial_moment_2,
#     }

#     for stat_name, stats in {
#         "local_bifurcation_angles": local_bifurcation_angles,
#         "remote_bifurcation_angles_stats": remote_bifurcation_angles,
#         "terminal_path_lengths": terminal_path_lengths,
#     }.items():
#         result.update(Statistics.gather_stats(stats, prefix=f"{stat_name}_"))

#     return result


# class ComputeStatistics(luigi_tools.task.WorkflowTask):
#     """Task to compute the statistics of the given morphologies."""

#     morph_dir = OptionalPathParameter(
#         description="Folder containing the input morphologies.",
#         default=None,
#     )
#     output_dataset = luigi.Parameter(description="Output dataset file", default="statistics.json")

#     def requires(self):
#         if self.morph_dir is None:
#             return RepairDataset()
#         return None

#     def run(self):
#         morph_dir = self.morph_dir or self.input().pathlib_path

#         pop = nm.core.Population(sorted(f for f in morph_dir.iterdir()))

#         result = population_statistics(pop)

#         with open(self.output().path, "w", encoding="utf-8") as f:
#             json.dump(result, f)

#         return result

#     def output(self):
#         return StatisticsOutputLocalTarget(self.output_dataset, create_parent=True)


# class PlotStatistics(luigi_tools.task.WorkflowTask):
#     """Task to plot the statistics of the given morphologies."""

#     output_dir = PathParameter(description="Output directory", default="figures")
#     nb_bins = luigi.IntParameter(description="The number of bins used for histograms", default=20)

#     def requires(self):
#         return ComputeStatistics()

#     def run(self):
#         with open(self.input().path, encoding="utf-8") as f:
#             statistics = json.load(f)

#         with PdfPages(self.output().pathlib_path / "input_statistics.pdf") as pdf:
#             for key, values in statistics.items():
#                 fig = plt.figure()
#                 ax = fig.gca()

#                 ax.hist(values, bins=self.nb_bins, density=True)

#                 ax.set_xlabel(key)
#                 ax.set_ylabel("Density")
#                 fig.suptitle(f"Input {key}")
#                 pdf.savefig()
#                 plt.close(fig)

#     def output(self):
#         return StatisticsOutputLocalTarget(self.output_dir, create_parent=True)


# class CompareStatistics(luigi_tools.task.WorkflowTask):
#     """Task to compare 2 sets of statistics."""

#     output_dir = PathParameter(description="Output directory", default="compare_statistics")
#     nb_bins = luigi.IntParameter(description="The number of bins used for histograms", default=20)
#     morph_dir_biological = OptionalPathParameter(
#         description="Folder containing the biological morphologies.",
#         default=None,
#     )
#     morph_dir_generated = OptionalPathParameter(
#         description="Folder containing the generated morphologies.",
#         default=None,
#     )

#     def requires(self):
#         return {
#             "bio": RepairDataset(),
#             "gen": AddTufts(),
#         }
#         # bio_kwargs = {"output_dataset": self.output_dir / "bio_stats.json"}
#         # if self.morph_dir_biological is not None:
#         #     bio_kwargs["morph_dir"] = self.morph_dir_biological
#         # else:
#         #     bio_kwargs["morph_dir"] = ClusterTerminals().output()["morphologies"].path
#         #     # bio_kwargs["morph_dir"] = RepairDataset().output().path

#         # gen_kwargs = {"output_dataset": self.output_dir / "gen_stats.json"}
#         # if self.morph_dir_generated is not None:
#         #     gen_kwargs["morph_dir"] = self.morph_dir_generated
#         # else:
#         #     gen_kwargs["morph_dir"] = SteinerMorphologies().output().path

#         # return {
#         #     "bio": ComputeStatistics(**bio_kwargs),
#         #     "gen": ComputeStatistics(**gen_kwargs),
#         # }

#     def run(self):
#         self.output().pathlib_path.mkdir(parents=True, exist_ok=True)

#         # import pdb
#         # pdb.set_trace()
#         plot_score_matrix(
#             self.input()["bio"].path,
#             self.input()["gen"]["morphologies_asc"].path,
#             self.output().pathlib_path / "score_matrix.pdf",
#             default_config(),
#         )

#         # with open(self.input()["bio"].path, encoding="utf-8") as f:
#         #     bio_statistics = json.load(f)
#         # with open(self.input()["gen"].path, encoding="utf-8") as f:
#         #     gen_statistics = json.load(f)

#         # with PdfPages(self.output().pathlib_path / "compare_statistics.pdf") as pdf:

#         #     for key, bio_values in bio_statistics.items():

#         #         gen_values = gen_statistics.get(key)

#         #         if gen_values is None:
#         #             logger.error(f"'{key}' was not found in {self.input()['gen'].path}")

#         #         fig = plt.figure()
#         #         ax = fig.gca()

#         #         gen_values = np.array(gen_values)
#         #         bio_values = np.array(bio_values)

#         #         values = gen_values / bio_values

#         #         ax.hist(values, bins=self.nb_bins, density=True)

#         #         ax.set_xlabel(f"Relative deviation for {key}")
#         #         ax.set_ylabel("Density")
#         #         fig.suptitle(f"Relative deviation for {key}")
#         #         pdf.savefig()
#         #         plt.close(fig)

#     def output(self):
#         return StatisticsOutputLocalTarget(self.output_dir)


# class CheckTuftStatistics(luigi_tools.task.WorkflowTask):
#     """Task to compute the statistics of the given tuft morphologies and check them."""

#     morph_dir = OptionalPathParameter(
#         description="Folder containing the input morphologies.",
#         default=None,
#     )
#     output_dataset = luigi.Parameter(description="Output dataset file", default="statistics.json")

#     def requires(self):
#         if self.morph_dir is None:
#             return ClusterTerminals()
#         return None

#     def find_morph_from_pt(self, tufts_by_region, x, y, z):
#         """Find a morphology from a given point."""
#         for region_name, tuft_files in tufts_by_region.items():
#             for i in tuft_files:
#                 morph = load_morphology(i)
#                 if (
#                     (morph.points[:, 0] >= x)
#                     & (morph.points[:, 0] < x + 1)
#                     & (morph.points[:, 1] >= y)
#                     & (morph.points[:, 1] < y + 1)
#                     & (morph.points[:, 2] >= z)
#                     & (morph.points[:, 2] < z + 1)
#                 ).any():
#                     print("File for:", x, y, z, "=>", region_name, i)

#     def run(self):
#         self.output().pathlib_path.mkdir(parents=True, exist_ok=True)
#         morph_dir = self.morph_dir or self.input()["tuft_morphologies"].pathlib_path
#         csv_files = sorted(filter(lambda x: x.suffix == ".csv", morph_dir.iterdir()))
#         tufts_by_region = defaultdict(list)
#         all_stats = {}
#         N = 0
#         for csv_file in csv_files:
#             df = pd.read_csv(csv_file)
#             df.drop(df.loc[df["tuft_morph_path"].isnull()].index, inplace=True)
#             for (region_name, _), tuft_files in df.groupby(["region_acronym", "cluster_id"]):
#                 tufts_by_region[region_name].append(tuft_files["tuft_morph_path"].tolist()[0])
#                 N += len(tuft_files["tuft_morph_path"].tolist())
#             if N > 100:
#                 break

#         for region_name, tuft_files in tufts_by_region.items():
#             all_stats[region_name] = neurots.extract_input.distributions(tuft_files)

#         self.find_morph_from_pt(tufts_by_region, 4435, 4137, 7273)
#         self.find_morph_from_pt(tufts_by_region, 4458, 5309, 7710)

#     def output(self):
#         return StatisticsOutputLocalTarget("tufts", create_parent=True)
